"""Testing the script create_rft_ertobs, both unit tests and integration tests"""
import datetime
import logging
import math
import os
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from fmu.tools.rms import create_rft_ertobs
from fmu.tools.rms.create_rft_ertobs import check_and_parse_config

logging.basicConfig(level=logging.INFO)

# Pylint exceptions for the mocked RMS object:
# pylint: disable=too-few-public-methods,no-self-use


class SurveyPointSeries:
    """Object representing all wellpaths in a gridmodel"""

    def __init__(self, wellname):
        self.wellname = wellname

    def get_measured_depths_and_points(self):
        """Return numpy array with coordinates/trajectory for individual wells.

        This mocked function can return some simple predefined wellname, and
        also load some trajectories from disk.

        Format::

            md, east, north, tvd

        """
        mock_coords = {"R-99": np.array([[0, 0, 0, 0], [2000, 0, 0, 2000]])}
        if self.wellname in mock_coords:
            return mock_coords[self.wellname]
        # Try to load coordinates from CSV on disk
        csvfilecandidate = (
            Path(__file__).parent
            / "rft_ertobs_data"
            / "coords"
            / ("coords-" + self.wellname + ".csv")
        )
        if csvfilecandidate.is_file():
            return np.array(pd.read_csv(csvfilecandidate, header=None))
        return None


class Trajectory:
    """Representing one trajectory for one well"""

    def __init__(self, wellname):
        self.wellname = wellname
        self.survey_point_series = SurveyPointSeries(wellname)


class Wellbore:
    """Represent "all" trajectories for one well"""

    def __init__(self, wellname):
        self.wellname = wellname
        self.trajectories = {
            # Add two trajectory name pointing to the same trajectories,
            # which trajectory is in use can be changed on a global basis.
            "Drilled trajectory": Trajectory(wellname),
            "Imported trajectory": Trajectory(wellname),
        }


class Well:
    """Represents a well, with multiple wellbores"""

    def __init__(self, wellname):
        self.wellname = wellname
        self.wellbore = Wellbore(wellname)


class Grid:
    """Represents the grid (geometry) in an RMS model"""

    def __init__(self):
        pass

    def get_cells_at_points(self, xyz):
        """Returns a cell index integer

        Args:
            xyz: List with three elements.

        Returns:
            integer being the cell index. nan if outside the grid.
        """
        # Simple mocked gridmodel, "three" cells, but with large indices.
        if xyz[0] > 1e8:
            return np.nan  # Outside grid
        if xyz[2] >= 1700:
            return 3000  # Volon
        if xyz[2] >= 1650:
            # Return as float, as this can be a later side effect
            # in pandas frames when other cells are np.nan
            return 2000.0  # Therys

        return 1000  # Valysar


class Property:
    """Represents a specific property for cells, here only zone
    values and names are mocked"""

    def __init__(self):
        self.code_names = {1: "Valysar", 2: "Therys", 3: "Volon"}

    def get_values(self):
        """Return a map from cell_index to zone_val"""
        # The mock assumes we have three grid cells, mapped to each zone.
        return {1000: 1, 2000: 2, 3000: 3}


class GridModel:
    """Represents a Gridmodel, that can associated to one grid."""

    def __init__(self):
        self.properties = {"Zone": Property()}

    def get_grid(self):
        """Get the associated grid"""
        return Grid()


class RMSMockedProject:
    """Represent the RMS/roxapi "project" magic variable"""

    def __init__(self):
        self.wells = {
            "R_A2": Well("R_A2"),
            "R_A3": Well("R_A3"),
            "R_A4": Well("R_A4"),
            "R_A5": Well("R_A5"),
            "R_A6": Well("R_A6"),
            "RFT_R_A2": Well("R_A2"),
            "RFT_R_A3": Well("R_A3"),
            "RFT_R_A4": Well("R_A4"),
            "RFT_R_A5": Well("R_A5"),
            "RFT_R_A6": Well("R_A6"),
            "RFT_55_33-A-2": Well("R_A2"),
            "RFT_55_33-A-3": Well("R_A3"),
            "RFT_55_33-A-4": Well("R_A4"),
            "RFT_55_33-A-5": Well("R_A5"),
            "RFT_55_33-A-6": Well("R_A6"),
            "R-99": Well("R-99"),
        }
        self.grid_models = {"Simgrid": GridModel()}


def test_get_well_coords():
    """Reliance on this test depends on whether the mocked
    RMS project resembles the real ROXAPI.
    """
    rms_project_mock = RMSMockedProject()

    assert (
        create_rft_ertobs.get_well_coords(rms_project_mock, "R-99")
        == np.array([[0, 0, 0, 0], [2000, 0, 0, 2000]])
    ).all()

    # Alternative trajectory name in RMS:
    assert (
        create_rft_ertobs.get_well_coords(
            rms_project_mock, "R-99", trajectory_name="Imported trajectory"
        )
        == np.array([[0, 0, 0, 0], [2000, 0, 0, 2000]])
    ).all()

    with pytest.raises(KeyError):
        create_rft_ertobs.get_well_coords(
            rms_project_mock, "R-99", trajectory_name="Bogus traj-name"
        )


@pytest.mark.parametrize(
    "coords, expected",
    [
        (np.array([[0, 0, 0, 0]]), True),
        (np.array([[0, 0, 0, 0], [1, 0, 0, 1]]), True),
        (np.array([[0, 0, 0, 0], [1, 0, 0, 1], [2, 0, 0, 1]]), False),
        (np.array([[0, 0, 0, 0], [1, 0, 0, 1], [2, 0, 0, 0]]), False),
        (np.array([[0, 0, 0, 0], [1, 0, 0, 1], [1.001, 0, 0, 0.999]]), False),
    ],
)
def test_strictly_downward(coords, expected):
    """Test that we can determine if wells go strictly downwards"""
    assert create_rft_ertobs.strictly_downward(coords) == expected


def test_interp_from_md():
    """Test interpolation along a wellpath from MD to XYZ"""
    coords = np.array([[0, 0, 0, 0], [1, 0, 0, 1]])
    assert create_rft_ertobs.interp_from_md(0.5, coords, interpolation="linear") == (
        0,
        0,
        0.5,
    )
    assert create_rft_ertobs.interp_from_md(0.5, coords, interpolation="cubic") == (
        0,
        0,
        0.5,
    )


def test_interp_from_xyz():
    """Test interpolation along a wellpath from XYZ to MD"""
    coords = np.array([[0, 0, 0, 0], [1, 0, 0, 1]])
    assert create_rft_ertobs.interp_from_xyz((0, 0, 0.5), coords) == 0.5

    # A wellpath going straight down, and then up again at 45 degrees:
    coords = np.array([[0, 0, 0, 0], [1, 0, 0, 1], [1 + math.sqrt(2), 1, 0, 0]])
    assert (
        create_rft_ertobs.interp_from_xyz((0, 0, 0.5), coords, interpolation="linear")
        == 0.35  # verify this number
    )
    assert (
        create_rft_ertobs.interp_from_xyz(
            (0.5, 0.0, 0.5), coords, interpolation="linear"
        )
        == 2.0  # verify this number
    )


def test_ertobs_df_to_files_1(tmpdir):
    """Test the writing of obs and txt files to disk, from a dataframe"""
    tmpdir.chdir()
    ertobs_df = pd.DataFrame(
        [
            {
                "WELL_NAME": "R-99",
                "DATE": datetime.date(2020, 6, 1),
                "EAST": 5555,
                "NORTH": 7777,
                "TVD": 2300,
                "MD": 2400,
                "ZONE": "Valyzar",
                "PRESSURE": 100,
                "ERROR": 3,
            }
        ]
    )
    create_rft_ertobs.ertobs_df_to_files(ertobs_df, tmpdir)
    assert Path("R-99.txt").read_text().strip() == "5555 7777 2400 2300 Valyzar"
    assert Path("R-99_1.obs").read_text().strip() == "100 3"
    assert Path("well_date_rft.txt").read_text().strip() == "R-99 01 06 2020 1"

    # Check that the file rft_ertobs.csv was created, this file
    # is for future use.
    pd.testing.assert_frame_equal(
        ertobs_df.astype(str),
        pd.read_csv(Path("rft_ertobs.csv")).astype(str),
    )


def test_main_mock_drogon(tmpdir):
    """Check that the code when executed on the Drogon case provides a
    predefined set of files (that has been manually verified)"""
    tmpdir.chdir()
    tmpdir.mkdir("exports")
    config = {
        "input_file": Path(__file__).parent / "rft_ertobs_data/input_table_drogon.csv",
        "exportdir": "exports",
        "project": RMSMockedProject(),
        "gridname": "Simgrid",
        "zonename": "Zone",
        "alias_file": Path(__file__).parent / "rft_ertobs_data/rms_eclipse.csv",
        "ecl_name": "ECLIPSE_WELL_NAME",
        "rms_name": "RMS_WELL_NAME",
    }
    create_rft_ertobs.main(config)

    expected_files_dir = Path(__file__).parent / "rft_ertobs_data" / "expected_files"
    for filename in (
        list(expected_files_dir.glob("R*obs"))
        + list(expected_files_dir.glob("R*.txt"))
        + [
            Path(__file__).parent
            / "rft_ertobs_data"
            / "expected_files"
            / "well_date_rft.txt"
        ]
    ):
        assert (Path("exports") / Path(filename)).is_file()
        pd.testing.assert_frame_equal(
            # Generated by code:
            pd.read_csv(
                Path("exports") / (Path(filename).name),
                sep=r"\s+",
                index_col=None,
                header=None,
            ),
            # Reference/expected data:
            pd.read_csv(filename, sep=r"\s+", index_col=None, header=None),
            check_dtype=False,
        )
    assert Path("exports/well_date_rft.txt").is_file()


def test_alternative_trajectory_name(tmpdir):
    """Test that we can use a different trajectory in RMS, but it has to be
    the same for all wells"""
    tmpdir.chdir()
    tmpdir.mkdir("exports")
    config = {
        "input_file": Path(__file__).parent / "rft_ertobs_data/input_table_drogon.csv",
        "exportdir": "exports",
        "project": RMSMockedProject(),
        "gridname": "Simgrid",
        "zonename": "Zone",
        "trajectory_name": "Imported trajectory",
    }
    create_rft_ertobs.main(config)
    assert not pd.read_csv(Path("exports") / "rft_ertobs.csv").empty

    # Check that we fail when the trajectory is wrongly named:
    config["trajectory_name"] = "fooo"
    with pytest.raises(KeyError):
        create_rft_ertobs.main(config)


def test_main_no_rms(tmpdir):
    """Test that if we have a full CSV file with all values we don't need
    the RMS project"""
    tmpdir.chdir()

    input_dframe = pd.DataFrame(
        data=[
            {
                "DATE": "2099-01-01",
                "WELL_NAME": "A-1",
                "MD": 4,
                "EAST": 1,
                "NORTH": 2,
                "TVD": 3,
                "PRESSURE": 100,
                "ERROR": 5,
            }
        ]
    )

    config = {"input_dframe": input_dframe}
    create_rft_ertobs.main(config)
    assert (
        pd.read_csv("A-1.txt", sep=r"\s+", index_col=None, header=None).iloc[0].values
        == [1, 2, 4, 3]
    ).all()
    assert (
        pd.read_csv("A-1_1.obs", sep=r"\s+", index_col=None, header=None).iloc[0].values
        == [100, 5]
    ).all()
    assert Path("well_date_rft.txt").read_text().strip() == "A-1 01 01 2099 1"


@pytest.mark.parametrize(
    "data, expected_error",
    [
        ({"PRESSURE": ""}, "PRESSURE not provided"),
        ({"PRESSURE": np.nan}, "PRESSURE not provided"),
        ({"ERROR": np.nan}, "ERROR not provided"),
        ({"ERROR": ""}, "ERROR not provided"),
    ],
)
def test_missing_dframe_data(data, expected_error, tmpdir):
    """Test error message when something is missing in the input dataframe"""
    tmpdir.chdir()
    dframe_dict = {
        "DATE": "2099-01-01",
        "WELL_NAME": "A-1",
        "MD": 4,
        "EAST": 1,
        "NORTH": 2,
        "TVD": 3,
        "PRESSURE": "100",
        "ERROR": 5,
    }
    dframe_dict.update(data)
    with pytest.raises(ValueError, match=expected_error):
        create_rft_ertobs.main({"input_dframe": pd.DataFrame(data=[dframe_dict])})


def test_absolute_error(tmpdir):
    """Test that we can specify absolute_error in the config"""
    tmpdir.chdir()
    dframe_dict = {
        "DATE": "2099-01-01",
        "WELL_NAME": "A-1",
        "MD": 4,
        "EAST": 1,
        "NORTH": 2,
        "TVD": 3,
        "PRESSURE": "100",
    }
    create_rft_ertobs.main(
        {"input_dframe": pd.DataFrame(data=[dframe_dict]), "absolute_error": 4}
    )
    assert (
        pd.read_csv("A-1_1.obs", sep=r"\s+", index_col=None, header=None).iloc[0].values
        == [100, 4]
    ).all()


def test_absolute_error_ignored(tmpdir):
    """Test that we can absolute_error in the config is ignored if ERROR is specified"""
    tmpdir.chdir()
    dframe_dict = {
        "DATE": "2099-01-01",
        "WELL_NAME": "A-1",
        "MD": 4,
        "EAST": 1,
        "NORTH": 2,
        "TVD": 3,
        "PRESSURE": "100",
        "ERROR": 5,
    }
    create_rft_ertobs.main(
        {"input_dframe": pd.DataFrame(data=[dframe_dict]), "absolute_error": 4}
    )
    assert (
        pd.read_csv("A-1_1.obs", sep=r"\s+", index_col=None, header=None).iloc[0].values
        == [100, 5]
    ).all()


def test_configparsing(tmpdir, caplog):
    """Test that the function that validates and parses the config dictionary
    gives correct error messages, and returns a dict with defaults filled in"""

    with pytest.raises(TypeError):
        # pylint: disable=no-value-for-parameter
        check_and_parse_config()
    with pytest.raises(AssertionError):
        check_and_parse_config({})

    with pytest.raises(ValueError, match="Missing columns"):
        check_and_parse_config({"input_dframe": pd.DataFrame()})

    minimal_config = {
        "input_dframe": pd.DataFrame(
            columns=["DATE", "MD", "WELL_NAME", "PRESSURE"],
            data=[[np.datetime64(datetime.date(2020, 3, 12)), 1000, "C-19", 222]],
        )
    }

    processed_df = check_and_parse_config(minimal_config)["input_dframe"]
    tmpdir.chdir()

    # Dump the dataframe to CSV and reload as disk dataframe:
    processed_df.to_csv("tmp.csv", index=False)
    viadisk_df = check_and_parse_config({"input_file": "tmp.csv"})["input_dframe"]
    pd.testing.assert_frame_equal(processed_df, viadisk_df, check_dtype=False)

    with pytest.raises(ValueError):
        check_and_parse_config(
            {"input_dframe": pd.DataFrame(), "input_file": "tmp.csv"}
        )

    check_and_parse_config({**minimal_config, **{"foobar": "com"}})
    assert "Unknown options ignored: {'foobar'}" in caplog.text

    assert check_and_parse_config(minimal_config)["exportdir"] == "."

    with pytest.raises(
        AssertionError, match="Only linear and cubic interpolation is supported"
    ):
        check_and_parse_config({**minimal_config, **{"interpolation": "bilinear"}})

    with pytest.raises(AssertionError, match="create upfront"):
        check_and_parse_config({**minimal_config, **{"exportdir": "nonexistingdir"}})

    assert check_and_parse_config(minimal_config)["welldatefile"] == "well_date_rft.txt"


def test_date_parsing(tmpdir):
    """Check that the ambiguous "DD MM YYYY" date format
    can be parsed "correctly" in input CSV files."""
    tmpdir.chdir()
    pd.DataFrame(
        columns=["DATE", "MD", "WELL_NAME", "PRESSURE"],
        data=[["01 03 2009", "100", "A-1", "123"]],
    ).to_csv("inputframe.csv", header=True, index=False)

    minimal_config = {"input_file": "inputframe.csv"}
    processed_config = check_and_parse_config(minimal_config)
    assert processed_config["input_dframe"]["DATE"].values[0] == np.datetime64(
        datetime.date(2009, 3, 1)
    )


def test_parse_alias_config(tmpdir):
    """Test that alias files can be parsed"""
    tmpdir.chdir()
    minimal_config = {
        "input_dframe": pd.DataFrame(columns=["DATE", "MD", "WELL_NAME", "PRESSURE"])
    }

    # Alias is defaulted to empty dict
    assert check_and_parse_config(minimal_config)["alias"] == {}

    # Empty alias dict ok to submit:
    check_and_parse_config({**minimal_config, **{"alias": {}}})

    assert (
        check_and_parse_config({**minimal_config, **{"alias": {"foo": "bar"}}})[
            "alias"
        ]["foo"]
        == "bar"
    )

    # First with no header
    alias_dframe = pd.DataFrame(data=[["NO 33/44 A-1", "A-1"], ["NO 33/44 A-2", "A-2"]])
    alias_dframe.to_csv("alias.csv", index=False)
    with pytest.raises(ValueError):
        # pylint: disable=expression-not-assigned
        check_and_parse_config({**minimal_config, **{"alias_file": "alias.csv"}})[
            "alias"
        ]

    # Use default header names
    alias_dframe = pd.DataFrame(
        columns=["RMS_WELL_NAME", "ECLIPSE_WELL_NAME"],
        data=[["NO 33/44 A-1", "A-1"], ["NO 33/44 A-2", "A-2"]],
    )
    alias_dframe.to_csv("alias.csv", index=False)
    assert (
        check_and_parse_config({**minimal_config, **{"alias_file": "alias.csv"}})[
            "alias"
        ]["NO 33/44 A-1"]
        == "A-1"
    )

    # Use custom header names for alias:
    alias_dframe = pd.DataFrame(
        columns=["RMS_NAME", "ECLIPSE_NAME"],
        data=[["NO 33/44 A-1", "A-1"], ["NO 33/44 A-2", "A-2"]],
    )
    alias_dframe.to_csv("alias.csv", index=False)
    assert (
        check_and_parse_config(
            {
                **minimal_config,
                **{
                    "alias_file": "alias.csv",
                    "rms_name": "RMS_NAME",
                    "ecl_name": "ECLIPSE_NAME",
                },
            }
        )["alias"]["NO 33/44 A-1"]
        == "A-1"
    )


@pytest.mark.parametrize("rftzone", [("Volon"), ("Nansen")])
def test_zones(rftzone, tmpdir, caplog):
    """Test that a warning is emitted for RFT observations that in the RMS grid
    has ended up in a different zone than the one requested in the input"""
    tmpdir.chdir()
    dframe_dict = {
        "DATE": "2099-01-01",
        "WELL_NAME": "R-99",
        "MD": 1900,
        "EAST": 0,
        "NORTH": 0,
        "TVD": 1900,
        "PRESSURE": "100",
        "ERROR": 5,
    }
    dframe_dict.update({"ZONE": rftzone})
    create_rft_ertobs.main(
        {
            "input_dframe": pd.DataFrame(data=[dframe_dict]),
            "project": RMSMockedProject(),
            "gridname": "Simgrid",
            "zonename": "Zone",
        }
    )
    # The zone in the RMS grid is set to Volon in the mocked RMS project.
    if rftzone == "Volon":
        assert "Some points are in a different zone in the RMS grid" not in caplog.text
    else:
        assert "Some points are in a different zone in the RMS grid" in caplog.text


def test_rft_outside_grid(tmp_path, caplog):
    """Test behaviour when points are outside the grid."""
    os.chdir(tmp_path)
    dframe_dict = {
        "DATE": "2099-01-01",
        "WELL_NAME": "R-99",
        "MD": 4,
        "EAST": 1e10,
        "NORTH": 2,
        "TVD": 3,
        "PRESSURE": "100",
        "ERROR": 5,
        "ZONE": "Volon",  # Presence of this give a warning when grid is not supplied
    }
    create_rft_ertobs.main({"input_dframe": pd.DataFrame(data=[dframe_dict])})

    assert "Cannot verify zones when no RMS project is provided" in caplog.text

    # No problems in this script.
    # These data points will be ignored later with GENDATA_RFT.
    assert (
        pd.read_csv("R-99.txt", sep=r"\s+", index_col=None, header=None).iloc[0].values
        == np.array([1e10, 2, 4, 3, "Volon"], dtype="object")
    ).all()


def test_rft_outside_grid_with_zone(tmp_path, caplog):
    """Test behaviour when points are outside the grid and we try
    to match with a zone (which can't be done when the point is outside the grid).
    """
    os.chdir(tmp_path)
    dframe_dict = {
        "DATE": "2099-01-01",
        "WELL_NAME": "R-99",
        "MD": 4,
        "EAST": 1e10,
        "NORTH": 2,
        "TVD": 3,
        "PRESSURE": "100",
        "ERROR": 5,
        "ZONE": "Volon",
    }

    create_rft_ertobs.main(
        {
            "input_dframe": pd.DataFrame(data=[dframe_dict]),
            "project": RMSMockedProject(),
            "gridname": "Simgrid",
            "zonename": "Zone",
        }
    )
    # The script only warns about the situation.
    assert "RFT points outside the grid" in caplog.text

    # Output files are still written with the data. GENDATA_RFT is responsible
    # for ignoring the observation as it will not be able to find the cell index for
    # it.
    assert (
        pd.read_csv("R-99.txt", sep=r"\s+", index_col=None, header=None).iloc[0].values
        == np.array([1e10, 2, 4, 3, "Volon"], dtype="object")
    ).all()


def test_report_step_same_xyz(tmpdir):
    """For wells with RFT observations at multiple dates, the REPORT_STEP
    parameter must be set to something unique, the current code enumerates it
    from 1 and up. This test is with a measurement repeated in the same point"""
    tmpdir.chdir()
    obs_1 = {
        "DATE": "2000-01-01",
        "WELL_NAME": "R-99",
        "MD": 1900,
        "EAST": 0,
        "NORTH": 0,
        "TVD": 1900,
        "PRESSURE": "100",  # (here we also test that strings go well)
        "ERROR": 5,
    }
    obs_2 = obs_1.copy()
    obs_2.update({"DATE": "2010-01-01", "PRESSURE": 90})
    create_rft_ertobs.main(
        {
            "input_dframe": pd.DataFrame(data=[obs_1, obs_2]),
        }
    )

    # RFT at multiple dates but at the same point in the reservoir should only have
    # one point in the trajectory file:
    pd.testing.assert_frame_equal(
        pd.read_csv(Path("R-99.txt"), sep=r"\s+", header=None),
        pd.DataFrame([[0, 0, 1900, 1900]]),
    )

    # But there must be multiple observation files, coupled with REPORT_STEP.
    print(pd.read_csv(Path("R-99_1.obs"), sep=r"\s+", header=None))
    pd.testing.assert_frame_equal(
        pd.read_csv(Path("R-99_1.obs"), sep=r"\s+", header=None),
        pd.DataFrame([[100, 5]]),
    )
    pd.testing.assert_frame_equal(
        pd.read_csv(Path("R-99_2.obs"), sep=r"\s+", header=None),
        pd.DataFrame([[90, 5]]),
    )

    # Check that we get REPORT_STEP 1 for the first date, and 2 for the second:
    pd.testing.assert_frame_equal(
        pd.read_csv(Path("well_date_rft.txt"), sep=r"\s+", header=None),
        pd.DataFrame([["R-99", 1, 1, 2000, 1], ["R-99", 1, 1, 2010, 2]]),
    )


def test_report_step_different_xyz(tmpdir):
    """Multiple dates for a well at different xyz.

    This requires a setup compatible with GENDATA_RFT in semeio and GEN_DATA in ERT.

    See also:
    https://github.com/equinor/semeio/blob/master/tests/jobs/rft/test_gendata_rft.py
    """
    tmpdir.chdir()
    obs_1 = {
        "DATE": "2000-01-01",
        "WELL_NAME": "R-99",
        "MD": 1900,
        "EAST": 0,
        "NORTH": 0,
        "TVD": 1900,
        "PRESSURE": "100",  # (here we also test that strings go well)
        "ERROR": 5,
    }
    obs_2 = obs_1.copy()
    obs_2.update({"DATE": "2010-01-01", "PRESSURE": 90, "TVD": 1920, "MD": 1920})
    create_rft_ertobs.main(
        {
            "input_dframe": pd.DataFrame(data=[obs_1, obs_2]),
        }
    )

    # Both points should be mentioned:
    pd.testing.assert_frame_equal(
        pd.read_csv(Path("R-99.txt"), sep=r"\s+", header=None),
        pd.DataFrame([[0, 0, 1900, 1900], [0, 0, 1920, 1920]]),
    )

    # But there must be multiple observation files, and points with no data
    # must be padded with -1, and filename coupled with REPORT_STEP.
    pd.testing.assert_frame_equal(
        pd.read_csv(Path("R-99_1.obs"), sep=r"\s+", header=None),
        pd.DataFrame([[100, 5.0], [-1, 0.0]]),
    )
    pd.testing.assert_frame_equal(
        pd.read_csv(Path("R-99_2.obs"), sep=r"\s+", header=None),
        pd.DataFrame([[-1, 0.0], [90, 5.0]]),
    )

    # Check that we get REPORT_STEP 1 for the first date, and 2 for the second:
    pd.testing.assert_frame_equal(
        pd.read_csv(Path("well_date_rft.txt"), sep=r"\s+", header=None),
        pd.DataFrame([["R-99", 1, 1, 2000, 1], ["R-99", 1, 1, 2010, 2]]),
    )



"""Testing generating design matrices from dictionary input"""

import shutil
from pathlib import Path
import subprocess
from packaging import version

import pandas as pd

import pytest

from fmu.tools.sensitivities import DesignMatrix

TESTDATA = Path(__file__).parent / "data"


def valid_designmatrix(dframe):
    """Performs general checks on a design matrix, that should always be valid"""
    assert "REAL" in dframe

    # REAL always starts at 0 and is consecutive
    assert dframe["REAL"][0] == 0
    assert dframe["REAL"].diff().dropna().unique() == 1

    assert "SENSNAME" in dframe.columns
    assert "SENSCASE" in dframe.columns

    # There should be no empty cells in the dataframe:
    assert not dframe.isna().sum().sum()


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_designmatrix():
    """Test the DesignMatrix class"""

    design = DesignMatrix()

    mock_dict = dict(
        designtype="onebyone",
        seeds="default",
        repeats=10,
        defaultvalues=dict(),
        sensitivities=dict(
            rms_seed=dict(seedname="RMS_SEED", senstype="seed", parameters=None)
        ),
    )

    design.generate(mock_dict)
    valid_designmatrix(design.designvalues)
    assert len(design.designvalues) == 10
    assert isinstance(design.defaultvalues, dict)


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_endpoint(tmpdir):
    """Test the installed endpoint

    Will write generated design matrices to the pytest tmpdir directory,
    usually /tmp/pytest-of-<username>/
    """
    designfile = TESTDATA / "config/design_input_onebyone.xlsx"

    # The xlsx file contains a relative path, relative to the input design sheet:
    dependency = (
        pd.read_excel(designfile, header=None, engine="openpyxl")
        .set_index([0])[1]
        .to_dict()["background"]
    )

    tmpdir.chdir()
    # Copy over input files:
    shutil.copy(str(designfile), ".")
    shutil.copy(Path(designfile).parent / dependency, ".")

    subprocess.run(["fmudesign", str(designfile)], check=True)

    assert Path("generateddesignmatrix.xlsx").exists  # Default output file
    valid_designmatrix(pd.read_excel("generateddesignmatrix.xlsx", engine="openpyxl"))

    subprocess.run(["fmudesign", str(designfile), "anotheroutput.xlsx"], check=True)
    assert Path("anotheroutput.xlsx").exists



"""Run tests in RMS.

Creates a tmp RMS project in given version which is used as fixture for all other Roxar
API dependent tests.

This requires a ROXAPI license, and to be ran in a "roxenvbash" environment; hence
the decorator "roxapilicense"

"""
from pathlib import Path
from os.path import isdir
import shutil
import numpy as np
import pandas as pd
import pytest

import xtgeo

try:
    import roxar
except ImportError:
    pass


# ======================================================================================
# settings to create RMS project!

TMPD = Path("TMP")
TMPD.mkdir(parents=True, exist_ok=True)

TPATH = Path("../xtgeo-testdata")
PROJNAME = "tmp_project.rmsxxx"

PRJ = str(TMPD / PROJNAME)

CUBEDATA1 = TPATH / "cubes/reek/syntseis_20000101_seismic_depth_stack.segy"
CUBENAME1 = "synth1"

SURFTOPS1 = [
    TPATH / "surfaces/reek/1/topreek_rota.gri",
    TPATH / "surfaces/reek/1/midreek_rota.gri",
    TPATH / "surfaces/reek/1/lowreek_rota.gri",
]

SURFCAT1 = "DS_whatever"
SURFNAMES1 = ["TopReek", "MidReek", "BaseReek"]

GRIDDATA1 = TPATH / "3dgrids/reek/reek_sim_grid.roff"
PORODATA1 = TPATH / "3dgrids/reek/reek_sim_poro.roff"
ZONEDATA1 = TPATH / "3dgrids/reek/reek_sim_zone.roff"
GRIDNAME1 = "Simgrid"
PORONAME1 = "PORO"
ZONENAME1 = "Zone"

WELLSFOLDER1 = TPATH / "wells/reek/1"
WELLS1 = ["OP1_perf.w", "OP_2.w", "OP_6.w", "XP_with_repeat.w"]


@pytest.mark.skipunlessroxar
@pytest.fixture(name="create_project", scope="module", autouse=True)
def fixture_create_project():
    """Create a tmp RMS project for testing, populate with basic data.

    After the yield command, the teardown phase will remove the tmp RMS project.
    """
    prj1 = str(PRJ)

    print("\n******** Setup RMS project!\n")
    if isdir(prj1):
        print("Remove existing project! (1)")
        shutil.rmtree(prj1)

    project = roxar.Project.create()

    rox = xtgeo.RoxUtils(project)
    print("Roxar version is", rox.roxversion)
    print("RMS version is", rox.rmsversion(rox.roxversion))
    assert "1." in rox.roxversion

    for wfile in WELLS1:
        wobj = xtgeo.well_from_file(WELLSFOLDER1 / wfile)
        if "XP_with" in wfile:
            wobj.name = "OP2_w_repeat"

        wobj.to_roxar(project, wobj.name, logrun="log", trajectory="My trajectory")

    # populate with cube data
    cube = xtgeo.cube_from_file(CUBEDATA1)
    cube.to_roxar(project, CUBENAME1, domain="depth")

    # populate with surface data
    rox.create_horizons_category(SURFCAT1)
    for num, name in enumerate(SURFNAMES1):
        srf = xtgeo.surface_from_file(SURFTOPS1[num])
        project.horizons.create(name, roxar.HorizonType.interpreted)
        srf.to_roxar(project, name, SURFCAT1)

    # populate with grid and props
    grd = xtgeo.grid_from_file(GRIDDATA1)
    grd.to_roxar(project, GRIDNAME1)
    por = xtgeo.gridproperty_from_file(PORODATA1, name=PORONAME1)
    por.to_roxar(project, GRIDNAME1, PORONAME1)
    zon = xtgeo.gridproperty_from_file(ZONEDATA1, name=ZONENAME1)
    zon.values = zon.values.astype(np.uint8)
    zon.to_roxar(project, GRIDNAME1, ZONENAME1)

    # save project (both an initla version and a work version) and exit
    project.save_as(prj1)
    project.close()

    yield project

    print("\n******* Teardown RMS project!\n")

    if isdir(prj1):
        print("Remove existing project! (1)")
        shutil.rmtree(prj1)


@pytest.mark.skipunlessroxar
def test_qcforward_wzonation_vs_grid_stops(tmp_path):
    """Test wzonation vs grid inside RMS, and here it will STOP."""
    # ==================================================================================
    # qcforward well vs grid (mimic python inside RMS input!)
    # pylint: disable=invalid-name
    from fmu.tools import qcforward

    PRJ1 = PRJ
    WELLS = ["OP.*"]

    ZONELOGNAME = "Zonelog"
    TRAJ = "My trajectory"
    LOGRUN = "log"
    GRIDNAME = "Simgrid"
    ZONEGRIDNAME = "Zone"
    DRANGE = [1300, 3200]
    ZLOGRANGE = [1, 3]
    ZLOGSHIFT = 0
    REPORTPATH = tmp_path / "well_vs_grid.csv"

    ACT = [
        {"warn": "any < 90", "stop": "any < 70"},
        {"warn": "all < 95", "stop": "all < 80"},
    ]

    qcjob = qcforward.WellZonationVsGrid()

    def check():

        usedata = {
            "wells": {"names": WELLS, "logrun": LOGRUN, "trajectory": TRAJ},
            "zonelog": {"name": ZONELOGNAME, "range": ZLOGRANGE, "shift": ZLOGSHIFT},
            "grid": GRIDNAME,
            "depthrange": DRANGE,
            "gridprops": [ZONEGRIDNAME],
            "actions": ACT,
            "report": REPORTPATH,
            "nametag": "ZONELOG",
        }

        qcjob.run(usedata, project=PRJ1)

    with pytest.raises(SystemExit):
        check()


@pytest.mark.skipunlessroxar
def test_qcforward_wzonation_vs_grid_runs_ok(tmp_path):
    """Test wzonation vs grid inside RMS, and here it will run OK."""
    # ==================================================================================
    # qcforward well vs grid (mimic python inside RMS input!)
    # pylint: disable=invalid-name
    from fmu.tools import qcforward

    PRJ1 = PRJ
    WELLS = ["OP.*"]

    ZONELOGNAME = "Zonelog"
    TRAJ = "My trajectory"
    LOGRUN = "log"
    GRIDNAME = "Simgrid"
    ZONEGRIDNAME = "Zone"
    DRANGE = [1300, 3200]
    ZLOGRANGE = [1, 3]
    ZLOGSHIFT = 0
    REPORTPATH = tmp_path / "well_vs_grid.csv"

    ACT = [
        {"warn": "any < 80", "stop": "any < 50"},
        {"warn": "all < 85", "stop": "all < 40"},
    ]

    qcjob = qcforward.WellZonationVsGrid()

    def check():

        usedata = {
            "wells": {"names": WELLS, "logrun": LOGRUN, "trajectory": TRAJ},
            "zonelog": {"name": ZONELOGNAME, "range": ZLOGRANGE, "shift": ZLOGSHIFT},
            "grid": GRIDNAME,
            "depthrange": DRANGE,
            "gridprops": [ZONEGRIDNAME],
            "actions": ACT,
            "report": REPORTPATH,
            "nametag": "ZONELOG",
        }

        qcjob.run(usedata, project=PRJ1)

    check()

    result = pd.read_csv(REPORTPATH).set_index("WELL")
    assert result.loc["OP_1_PERF", "MATCH%"] == pytest.approx(70.588235)
    assert result.loc["all", "MATCH%"] == pytest.approx(67.207573)


@pytest.mark.skipunlessroxar
def test_qcforward_gridquality_ok(tmp_path):
    """Test qcforward gridquality parameters that runs ok."""
    # ==================================================================================
    # qcforward grid quality (mimic python inside RMS input!)
    # pylint: disable=invalid-name
    from fmu.tools import qcforward

    PRJ1 = PRJ
    GRIDNAME = "Simgrid"
    REPORT = tmp_path / "gridquality.csv"

    ACT = {
        "minangle_topbase": [
            {"warn": "allcells > 1% when < 80", "stop": "allcells > 1% when < 50"},
            {"warn": "allcells > 50% when < 85", "stop": "all > 10% when < 50"},
            {"warn": "allcells > 50% when < 85"},
        ],
        "collapsed": [{"warn": "all > 20%", "stop": "all > 50%"}],
        "faulted": [{"warn": "all > 20%", "stop": "all > 50%"}],
    }

    qcjob = qcforward.GridQuality()

    def check():

        usedata = {
            "grid": GRIDNAME,
            "actions": ACT,
            "report": {"file": REPORT, "mode": "write"},
            "nametag": "ZONELOG",
        }

        qcjob.run(usedata, project=PRJ1)

    check()

    result = pd.read_csv(REPORT).set_index("GRIDQUALITY")
    assert result.loc["minangle_topbase[1]", "WARN%"] == pytest.approx(10.447, abs=0.01)
    assert result.loc["minangle_topbase[0]", "WARN%"] == pytest.approx(2.715, abs=0.01)


@pytest.mark.skipunlessroxar
def test_qcforward_gridquality_fail(tmp_path):
    """Test qcforward gridquality parameters that shall fail on faulted."""
    # ==================================================================================
    # qcforward grid quality (mimic python inside RMS input!)
    # pylint: disable=invalid-name
    from fmu.tools import qcforward

    PRJ1 = PRJ
    GRIDNAME = "Simgrid"
    REPORT = tmp_path / "gridquality.csv"

    ACT = {
        "minangle_topbase": [
            {"warn": "allcells > 1% when < 80", "stop": "allcells > 1% when < 50"},
            {"warn": "allcells > 50% when < 85", "stop": "all > 10% when < 50"},
            {"warn": "allcells > 50% when < 85"},
        ],
        "collapsed": [{"warn": "all > 20%", "stop": "all > 50%"}],
        "faulted": [{"warn": "all > 20%", "stop": "all > 10%"}],
    }

    qcjob = qcforward.GridQuality()

    def check():

        usedata = {
            "grid": GRIDNAME,
            "actions": ACT,
            "report": {"file": REPORT, "mode": "write"},
            "nametag": "ZONELOG",
        }

        qcjob.run(usedata, project=PRJ1)

    with pytest.raises(SystemExit):
        check()

    result = pd.read_csv(REPORT).set_index("GRIDQUALITY")
    assert result.loc["faulted[0]", "WARN%"] == pytest.approx(16.368, abs=0.01)



from typing import Optional, List, Any
import pandas as pd

from fmu.tools.qcdata import QCData
from fmu.tools._common import _QCCommon
from fmu.tools.qcproperties._config_parser import ConfigParser
from fmu.tools.qcproperties._utils import filter_df

QCC = _QCCommon()


class WellLogs2df:
    """
    Class responsible for generating a property dataframe from well logs, and
    providing control arguments for the statisics extraction using PropertyAggregation()
    """

    def __init__(
        self,
        project: Optional[object],
        data: dict,
        xtgdata: QCData,
        blockedwells: bool = False,
    ):

        """Initiate instance"""
        QCC.verbosity = data.get("verbosity", 0)

        self._xtgdata = xtgdata  # A QCData instance used for dataloading to XTGeo
        self._wells: List[Any] = []
        self._property_type: Optional[str] = None
        self._dataframe = pd.DataFrame()  # dataframe with property log data

        self._data_input_preparations(project, data, blockedwells)

        # Get dataframe from the XTGeo objects
        self._create_df_from_wells()

    # ==================================================================================
    # Class properties
    # ==================================================================================

    @property
    def dataframe(self) -> pd.DataFrame:
        """Dataframe with property statistics."""
        return self._dataframe

    @property
    def property_type(self) -> Optional[str]:
        """Property type (continous/discrete)"""
        return self._property_type

    @property
    def aggregation_controls(self) -> dict:
        """Attribute to use for statistics aggregation"""
        return self._aggregation_controls

    # ==================================================================================
    # Hidden class methods
    # ==================================================================================

    def _data_input_preparations(
        self, project: Optional[object], data: dict, blockedwells: bool
    ):
        """
        Prepare the input parameter data for usage within QCProperties().
        Parameters are loaded to XTGeo and property types are checked.
        """
        data = data.copy()

        if blockedwells:
            data["bwells"] = data.pop("wells")

        controllers = ConfigParser(data)

        self._aggregation_controls = controllers.aggregation_controls
        self._controls = controllers.prop2df_controls

        # Load data to XTGeo
        self._xtgdata.parse(
            project=project,
            data=controllers.data_loading_input,
            reuse=True,
            wells_settings={"lognames": self._controls["unique_parameters"]},
        )

        self._set_wells(blockedwells)

        # Check which property type is input
        self._check_logs_and_set_property_type()

    def _set_wells(self, blockedwells: bool):
        """Set wells attribute"""
        self._wells = (
            self._xtgdata.wells.wells
            if not blockedwells
            else self._xtgdata.bwells.wells
        )
        self._validate_wells()

    def _validate_wells(self):
        """Remove wells where selector logs are missing"""
        selectors = self._controls["selectors_input_names"]
        removed_wells = []
        for xtg_well in self._wells:
            # skip well if selector logs are missing
            if not all(log in xtg_well.lognames for log in selectors):
                QCC.print_info(
                    f"Skipping {xtg_well.name} some selector logs are missing"
                )
                removed_wells.append(xtg_well)
                continue
        self._wells = [
            xtg_well for xtg_well in self._wells if xtg_well not in removed_wells
        ]

    def _check_logs_and_set_property_type(self):
        """
        Use XTGeo to check that selectors are discrete, and also
        check if input properties are continous or discrete.
        Raise errors if not desired format.
        """
        # check that all selectors are discrete
        selectors = self._controls["selectors_input_names"]
        if not all(self._wells[0].isdiscrete(log) for log in selectors):
            raise ValueError("Only discrete logs can be used as selectors")

        # check that all properties defined are of the same type
        properties = self._controls["properties_input_names"]
        if any(self._wells[0].isdiscrete(log) for log in properties) and not all(
            self._wells[0].isdiscrete(log) for log in properties
        ):
            raise TypeError(
                "Properties of different types (continuous/discrete) "
                "defined in the input."
            )

        # Set attribute used to control aggregation method
        discrete = self._wells[0].isdiscrete(properties[0])
        QCC.print_debug(
            f"{'Discrete' if discrete else 'Continous'} properties in input"
        )
        self._property_type = "DISC" if discrete else "CONT"

    def _codes_to_codenames(self):
        """Replace codes in dicrete parameters with codenames"""
        for param in self._controls["unique_parameters"]:

            if self._wells[0].isdiscrete(param):
                codes = self._wells[0].get_logrecord(param).copy()
                usercodes = self._controls["usercodes"].copy()

                # Update code names if user input
                if usercodes and param in usercodes:
                    codes.update(usercodes[param])

                # replace codes values in dataframe with code names
                self._dataframe[param] = self._dataframe[param].map(codes.get)

    def _create_df_from_wells(self):
        """
        Create a combined property dataframe for the input wells.
        Values for discrete logs will be replaced by their codename.
        """
        QCC.print_info("Creating property dataframe from well logs")
        # Loop through XTGeo wells and combine into one dataframe
        dfs = []
        for xtg_well in self._wells:
            # extract dataframe for well
            df_well = xtg_well.dataframe.copy()
            df_well["WELL"] = xtg_well.name
            dfs.append(df_well)

        dframe = pd.concat(dfs)

        # To avoid bias in statistics, drop duplicates to remove
        # cells penetrated by multiple wells.
        dframe = dframe.drop_duplicates(
            subset=[x for x in dframe.columns if x != "WELL"]
        )
        self._dataframe = dframe[self._controls["unique_parameters"]].copy()

        # replace codes values in dataframe with code names
        self._codes_to_codenames()

        # Filter property dataframe
        if self._controls["filters"]:
            self._dataframe = filter_df(self._dataframe, self._controls["filters"])

        # rename columns in dataframe
        self.dataframe.rename(columns=self._controls["name_mapping"], inplace=True)



""" Module for summarizing design set up for one by one sensitivities """
import pandas as pd


def summarize_design(filename, sheetname="DesignSheet01"):
    """
     Summarizes the design set up for one by one sensitivities
     specified in a design matrix on standard fmu format.

    Args:
        filename (str): Path to excel or csv file containting designmatrix
            for one by one sensitivities on standard FMU format.
        sheetname (str): Name of sheet in excel workbook which
            contains the designmatrix (only for excel input). Defaults to
            'DesignSheet01'.

    Returns:
        pd.DataFrame: Summary of sensitivities,
        corresponding realisation numbers,
        senstype('mc' or 'scalar')
        and senscase (name of high and low cases).
        Each row represents one sensitivity
        with 1-2 cases (low/high).
        Column names are ['sensno', 'sensname',
        'senstype', 'casename1', 'startreal1', 'endreal1',
        'casename2', 'startreal2', 'endreal2']

    Example::
        >>> from fmu.tools.sensitivities import summarize_design
        >>> designname = 'design_filename.xlsx'
        >>> designsheet = 'DesignSheet01'
        >>> designtable = summarize_design(designname, designsheet)

    """

    # Initialisation of dataframe to store results
    designsummary = pd.DataFrame(
        columns=[
            "sensno",
            "sensname",
            "senstype",
            "casename1",
            "startreal1",
            "endreal1",
            "casename2",
            "startreal2",
            "endreal2",
        ]
    )
    sensno = 0
    startreal1 = 0
    endreal1 = 0

    # Read design matrix and find realisation numbers for each sensitivity
    if str(filename).endswith(".xlsx"):
        dgn = pd.read_excel(filename, sheetname, engine="openpyxl")

        # Drop empty rows or columns that have been read in
        # due to having background colour/formatting

        dgn.dropna(axis=0, how="all", inplace=True)
        dgn = dgn.loc[:, ~dgn.columns.str.contains("^Unnamed")]

    elif str(filename).endswith(".csv"):
        dgn = pd.read_csv(filename)

    else:
        raise ValueError(
            "Design matrix must be on Excel or csv format"
            " and filename must end with .xlsx or .csv"
        )
    sensname = dgn.loc[0]["SENSNAME"]
    casename1 = dgn.loc[0]["SENSCASE"]
    if casename1.lower() == "p10_p90":
        senstype = "mc"
    elif casename1.lower() == "ref":
        senstype = "ref"
    else:
        senstype = "scalar"

    currentsensname = sensname
    currentsenscase = casename1
    # starting with first case
    secondcase = False
    casename2 = None
    startreal2 = None
    endreal2 = None

    for row in dgn.itertuples():
        if row.SENSNAME == currentsensname and row.SENSCASE == currentsenscase:
            if secondcase is True:
                endreal2 = row.REAL
            else:
                endreal1 = row.REAL
        elif row.SENSNAME == currentsensname:
            secondcase = True
            startreal2 = row.REAL
            endreal2 = row.REAL
            casename2 = row.SENSCASE
            currentsensname = row.SENSNAME
            currentsenscase = casename2
        else:
            if senstype != "skip":
                if secondcase is True:
                    designsummary.loc[sensno] = [
                        sensno,
                        sensname,
                        senstype,
                        casename1,
                        startreal1,
                        endreal1,
                        casename2,
                        startreal2,
                        endreal2,
                    ]
                    sensno += 1
                else:
                    designsummary.loc[sensno] = [
                        sensno,
                        sensname,
                        senstype,
                        casename1,
                        startreal1,
                        endreal1,
                        None,
                        None,
                        None,
                    ]
                    sensno += 1
            secondcase = False
            startreal1 = row.REAL
            endreal1 = row.REAL

            casename1 = row.SENSCASE
            sensname = row.SENSNAME
            currentsenscase = casename1
            currentsensname = sensname
            if row.SENSCASE.lower() == "p10_p90":
                senstype = "mc"
            elif row.SENSCASE.lower() == "skip":
                senstype = "skip"
            else:
                senstype = "scalar"

    # For last row
    if senstype != "skip":
        if secondcase is True:
            designsummary.loc[sensno] = [
                sensno,
                sensname,
                senstype,
                casename1,
                startreal1,
                endreal1,
                casename2,
                startreal2,
                endreal2,
            ]
        else:
            designsummary.loc[sensno] = [
                sensno,
                sensname,
                senstype,
                casename1,
                startreal1,
                endreal1,
                None,
                None,
                None,
            ]

    return designsummary



"""Module containing ....  """
import pandas as pd
import numpy as np

from fmu.tools._common import _QCCommon

from fmu.tools.qcproperties._utils import list_combinations

QCC = _QCCommon()


class PropertyAggregation:
    """
    Class for extracting statistics from a property dataframe.
    Statistics for multiple properties can be calculated simultaneosly. The
    aggregation methods and statistics are based on the property type.
    Selectors can be used to extract statistics per value in discrete properties.
    """

    def __init__(
        self,
        props2df,
    ):

        """Initiate instance"""
        self._property_dataframe = props2df.dataframe  # dataframe with properties
        self._controls = props2df.aggregation_controls
        self._controls["property_type"] = props2df.property_type
        self._dataframe = pd.DataFrame()  # dataframe with statistics

        QCC.verbosity = self._controls["verbosity"]

        # Create list with groups for pandas groupby
        selector_combo_list = (
            list_combinations(self._controls["selectors"])
            if self._controls["selector_combos"]
            else [self._controls["selectors"]]
        )

        # Compute dataframe with statistics
        QCC.print_info(
            f"Calculating statistics for {self._controls['property_type']} properties"
        )
        self._dataframe = (
            self._calculate_continous_statistics(selector_combo_list)
            if self._controls["property_type"] == "CONT"
            else self._calculate_discrete_fractions(selector_combo_list)
        )

    # ==================================================================================
    # Class properties
    # ==================================================================================

    @property
    def dataframe(self):
        """Returns the dataframe with property statistics."""
        return self._dataframe

    @property
    def controls(self):
        """Attribute with data used for statistics aggregation."""
        return self._controls

    # ==================================================================================
    # Hidden class methods
    # ==================================================================================

    def _disc_aggregations(self):
        """Statistical aggregations to extract from continous data"""
        return [
            ("Count", "count"),
            (
                "Sum_Weight",
                lambda x: np.sum(x)
                if x.name in list(self._controls["weights"].values())
                else np.nan,
            ),
        ]

    def _cont_aggregations(self, dframe=None):
        """Statistical aggregations to extract from discrete data"""
        return [
            ("Avg", np.mean),
            ("Stddev", np.std),
            ("P10", lambda x: np.nanpercentile(x, q=10)),
            ("P90", lambda x: np.nanpercentile(x, q=90)),
            ("Min", np.min),
            ("Max", np.max),
            (
                "Avg_Weighted",
                lambda x: np.average(
                    x.dropna(),
                    weights=dframe.loc[
                        x.dropna().index, self._controls["weights"][x.name]
                    ],
                )
                if x.name in self._controls["weights"]
                else np.nan,
            ),
            ("Count", "count"),
        ]

    def _calculate_continous_statistics(self, selector_combo_list):
        """
        Calculate statistics for continous properties.
        Returns a pandas dataframe.
        """

        # Extract statistics for combinations of selectors
        dfs = []
        groups = []
        for combo in selector_combo_list:
            group = self._property_dataframe.dropna(subset=combo).groupby(combo)
            groups.append(group)

            df_group = (
                group[self._controls["properties"]]
                .agg(self._cont_aggregations(dframe=self._property_dataframe))
                .stack(0)
                .rename_axis(combo + ["PROPERTY"])
                .reset_index()
            )
            dfs.append(df_group)

        # Extract statistics for the total
        group_total = self._property_dataframe.dropna(
            subset=self._controls["selectors"]
        ).groupby(lambda x: "Total")
        df_group = (
            group_total[self._controls["properties"]]
            .agg(self._cont_aggregations(dframe=self._property_dataframe))
            .stack(0)
            .reset_index(level=0, drop=True)
            .rename_axis(["PROPERTY"])
            .reset_index()
        )
        dfs.append(df_group)
        dframe = pd.concat(dfs)

        # Empty values in selectors is filled with "Total"
        dframe[self._controls["selectors"]] = dframe[
            self._controls["selectors"]
        ].fillna("Total")

        return dframe

    def _calculate_discrete_fractions(self, selector_combo_list):
        """
        Calculate fraction statistics for discrete properties. A Weighted
        fraction is calculated for each property where a weight is specified
        Returns a pandas dataframe.
        """

        # Extract statistics for combinations of selectors
        combo_list = selector_combo_list
        dfs = []
        for prop in self._controls["properties"]:
            if prop not in self._controls["selectors"]:
                combo_list = [x + [prop] for x in selector_combo_list]
                combo_list.append([prop])
                self._controls["selectors"].append(prop)

            select = (
                self._controls["weights"][prop]
                if prop in self._controls["weights"]
                else prop
            )

            for combo in combo_list:
                df_prop = self._property_dataframe.dropna(subset=combo).copy()
                df_group = (
                    df_prop.groupby(combo)[select]
                    .agg(self._disc_aggregations())
                    .reset_index()
                    .assign(PROPERTY=prop)
                )

                for col, name in {
                    "Avg_Weighted": "Sum_Weight",
                    "Avg": "Count",
                }.items():
                    df_group[f"Total_{name}"] = (
                        df_group.groupby([x for x in combo if x != prop])[
                            name
                        ].transform(lambda x: x.sum())
                        if combo != [prop]
                        else df_group[name].sum()
                    )
                    df_group[col] = df_group[name] / df_group[f"Total_{name}"]

                df_group = df_group.drop(
                    columns=["Total_Sum_Weight", "Total_Count", "Sum_Weight"]
                )
                dfs.append(df_group)

        dframe = pd.concat(dfs)

        # Empty values in selectors is filled with "Total"
        dframe[self._controls["selectors"]] = dframe[
            self._controls["selectors"]
        ].fillna("Total")

        return dframe



"""Testing qcforward method blockedwells vs gridprops"""

from copy import deepcopy
from os.path import abspath

import pandas as pd
import pytest
from fmu.tools import qcforward as qcf

# filedata
PATH = abspath(".")  # normally not needed; here due to pytest fixture tmpdir
GRIDFILE = "../xtgeo-testdata/3dgrids/drogon/3/valysar.roff"
BWELLFILES = [
    "../xtgeo-testdata/wells/drogon/3/valysar*.bw",
]
COMPARE = {"Facies": "FACIES", "PHIT": "PHIT"}


@pytest.fixture(name="datainput")
def fixture_datainput(tmp_path):
    data = {
        "nametag": "MYDATA1",
        "verbosity": "info",
        "path": PATH,
        "grid": GRIDFILE,
        "report": tmp_path / "bw_vs_gprop.csv",
        "compare": COMPARE,
        "gridprops": [["FACIES", GRIDFILE], ["PHIT", GRIDFILE]],
        "bwells": BWELLFILES,
        "tolerance": 0.01,
        "actions": [
            {"warn": "anywell < 80%", "stop": "anywell < 75%"},
            {"warn": "allwells < 90%", "stop": "allwells < 85%"},
        ],
    }
    return data


def test_bw_vs_gridprops_asfiles(datainput):
    """Testing the zonelog vs grid functionality using files"""

    job = qcf.BlockedWellsVsGridProperties()
    job.run(datainput)

    rep = pd.read_csv(datainput["report"])

    # pylint: disable=no-member, unsubscriptable-object
    assert rep.iloc[1].at["MATCH%"] == 85.0
    assert rep.iloc[1].at["STATUS"] == "OK"

    wel = "WELL"
    cmp = "COMPARE(BW:MODEL)"

    ser = rep.loc[(rep[wel] == "55_33-A-6") & (rep[cmp] == "PHIT:PHIT"), "MATCH%"]
    ser.reset_index(drop=True, inplace=True)
    assert ser[0] == 100.0


def test_bw_vs_gridprops_asfiles_shall_stop(tmp_path, datainput):
    """Testing the zonelog vs grid functionality using files, shall stop."""

    job = qcf.BlockedWellsVsGridProperties()
    data = deepcopy(datainput)
    data["actions"] = [
        {"warn": "anywell < 80%", "stop": "anywell < 86%"},
        {"warn": "allwells < 98%", "stop": "allwells < 96%"},
    ]
    data["report"] = tmp_path / "failedreport.csv"

    with pytest.raises(SystemExit) as err:
        job.run(data)
    assert "STOP criteria is found" in str(err)
    rep = pd.read_csv(data["report"])
    # pylint: disable=no-member, unsubscriptable-object
    assert rep.iloc[1].at["MATCH%"] == 85.0
    assert rep.iloc[1].at["STATUS"] == "STOP"


def test_bw_vs_gridprops_asfiles_change_tolerance(tmp_path, datainput):
    """Testing the zonelog vs grid functionality using files, iterate tolerance."""

    job = qcf.BlockedWellsVsGridProperties()
    data = deepcopy(datainput)
    data["actions"] = [
        {"warn": "anywell < 88%", "stop": "anywell < 86%"},
        {"warn": "allwells < 98%", "stop": "allwells < 98%"},
    ]
    data["report"] = tmp_path / "failedreport2.csv"
    data["tolerance"] = {"abs": 0.001}

    with pytest.raises(SystemExit) as err:
        job.run(data)
    assert "STOP criteria is found" in str(err)
    rep = pd.read_csv(data["report"])
    # pylint: disable=no-member, unsubscriptable-object
    assert rep.iloc[1].at["MATCH%"] == 85.0
    assert rep.iloc[1].at["STATUS"] == "STOP"



from typing import Optional
import pandas as pd
from fmu.tools.qcdata import QCData
from fmu.tools._common import _QCCommon
from fmu.tools.qcproperties._config_parser import ConfigParser
from fmu.tools.qcproperties._utils import filter_df

QCC = _QCCommon()


class GridProps2df:
    """
    Class responsible for generating a property dataframe from grid prperties, and
    providing control arguments for the statisics extraction using PropertyAggregation()
    """

    def __init__(self, project: Optional[object], data: dict, xtgdata: QCData):

        """Initiate instance"""
        QCC.verbosity = data.get("verbosity", 0)

        self._xtgdata = xtgdata  # A QCData instance used for dataloading to XTGeo
        self._property_type: Optional[str] = None
        self._dataframe = pd.DataFrame()  # dataframe with property data

        self._data_input_preparations(project, data)

        # Create dataframe from grid properties
        self._create_df_from_grid_props()

    # ==================================================================================
    # Class properties
    # ==================================================================================

    @property
    def dataframe(self) -> pd.DataFrame:
        """Dataframe with property statistics."""
        return self._dataframe

    @property
    def property_type(self) -> Optional[str]:
        """Property type (continous/discrete)"""
        return self._property_type

    @property
    def aggregation_controls(self) -> dict:
        """Attribute to use for statistics aggregation"""
        return self._aggregation_controls

    # ==================================================================================
    # Hidden class methods
    # ==================================================================================

    def _data_input_preparations(self, project: Optional[object], data: dict):
        """
        Prepare the input parameter data for usage within QCProperties().
        Parameters are loaded to XTGeo and property types are checked.
        """
        data = data.copy()
        controllers = ConfigParser(data)

        self._aggregation_controls = controllers.aggregation_controls
        self._controls = controllers.prop2df_controls

        xtg_input = controllers.data_loading_input

        # set gridprops argument format for dataloading with QCData()
        if not xtg_input["pfiles"]:
            xtg_input["gridprops"] = self._controls["unique_parameters"]
        else:
            xtg_input["gridprops"] = [
                [param, xtg_input["pfiles"][param]]
                if param in xtg_input["pfiles"]
                else param
                for param in self._controls["unique_parameters"]
            ]

        # Load data to XTGeo
        self._xtgdata.parse(
            project=project,
            data=xtg_input,
            reuse=True,
        )
        # Load data to XTGeo
        self._check_and_set_property_type()

    def _create_df_from_grid_props(self):
        """
        Extract a combined property dataframe for the input properties.
        Values for discrete logs will be replaced by their codename.
        """
        QCC.print_info("Creating property dataframe from grid properties")

        self._dataframe = self._xtgdata.gridprops.get_dataframe().copy().dropna()

        # replace codes values in dataframe with code names
        self._codes_to_codenames()

        # Filter property dataframe
        if self._controls["filters"]:
            self._dataframe = filter_df(self._dataframe, self._controls["filters"])

        # rename columns in dataframe
        self.dataframe.rename(columns=self._controls["name_mapping"], inplace=True)

    def _check_and_set_property_type(self):
        """
        Use XTGeo to check that selectors are discrete, and also
        check if input properties are continous or discrete.
        Raise errors if not desired format.
        """
        # check that all selectors are discrete
        selectors = self._controls["selectors_input_names"]
        xtgprops = [
            self._xtgdata.gridprops.get_prop_by_name(prop) for prop in selectors
        ]
        if not all(prop.isdiscrete for prop in xtgprops):
            raise ValueError("Only discrete properties can be used as selectors")

        # check that all properties defined are of the same type
        properties = self._controls["properties_input_names"]
        xtgprops = [
            self._xtgdata.gridprops.get_prop_by_name(prop) for prop in properties
        ]
        if any(prop.isdiscrete for prop in xtgprops) and not all(
            prop.isdiscrete for prop in xtgprops
        ):
            raise TypeError(
                "Properties of different types (continuous/discrete) "
                "defined in the input."
            )

        # Set attribute used to control aggregation method
        discrete = xtgprops[0].isdiscrete
        QCC.print_debug(
            f"{'Discrete' if discrete else 'Continous'} properties in input"
        )
        self._property_type = "DISC" if discrete else "CONT"

    def _codes_to_codenames(self):
        """Replace codes in dicrete parameters with codenames"""
        for param in self._controls["unique_parameters"]:
            xtg_prop = self._xtgdata.gridprops.get_prop_by_name(param)

            if xtg_prop.isdiscrete:
                codes = xtg_prop.codes.copy()
                usercodes = self._controls["usercodes"].copy()

                # Update code names if user input
                if usercodes and param in usercodes:
                    codes.update(usercodes[param])

                # replace codes values in dataframe with code names
                self._dataframe[param] = self._dataframe[param].map(codes.get)



"""
This private module in qcforward is used to check wellzonation vs grid zonation
"""

import collections
import json
from pathlib import Path

import fmu.tools
import numpy as np
from fmu.tools._common import _QCCommon
from fmu.tools.qcforward._qcforward import ActionsParser, QCForward
from jsonschema import validate

QCC = _QCCommon()
UNDEF = float("nan")


class _LocalData:
    def __init__(self):
        """Defining and hold data local for this routine"""

        self.zonelogname = "Zonelog"
        self.zonelogrange = [0, 99]
        self.zonelogshift = 0
        self.depthrange = [0.0, 999999.0]
        self.actions = None
        self.perflogname = None
        self.perflogrange = [1, 9999]
        self.gridzone = None
        self.gridzonerange = [1, 9999]
        self.wellresample = None
        self.infotext = "ZONELOG MATCH"
        self.nametag = None
        self.reportfile = None

    def parse_data(self, data):
        """Parsing the actual data"""

        # TODO: verify and qc
        self.nametag = data.get("nametag", "unset_nametag")
        if "report" in data:
            self.reportfile = (
                data["report"].get("file")
                if isinstance(data["report"], dict)
                else data["report"]
            )

        self.zonelogname = data["zonelog"]["name"]
        self.zonelogrange = data["zonelog"]["range"]
        self.zonelogshift = data["zonelog"].get("shift", 0)

        self.depthrange = data.get("depthrange", [0.0, 99999.0])

        self.actions = data["actions"]

        if "perflog" in data.keys() and data["perflog"]:
            self.perflogname = data["perflog"].get("name", None)
            self.perflogrange = data["perflog"].get("range", [1, 9999])
            self.infotext = "PERFLOG MATCH"

        if "well_resample" in data.keys():
            self.wellresample = data.get("well_resample", None)


class WellZonationVsGrid(QCForward):
    def run(self, data, reuse=False, project=None):
        """Main routine for evaulating well zonation match in 3D grids.

        The routine depends on existing XTGeo functions for this purpose

        Args:
            data (dict or str): The input data either as a Python dictionary or
                a path to a YAML file
            reuse (bool or list): Reusing some "timeconsuming to read" data in the
                instance. If True, then grid and gridprops will be reused as default.
                Alternatively it can be a list for more fine grained control, e.g.
                ["grid", "gridprops", "wells"]
            project (Union[object, str]): For usage inside RMS

        """
        data = self._data = self.handle_data(data, project)
        self._validate_input(self._data, project)

        QCC.verbosity = data.get("verbosity", 0)

        # parse data that are special (local) for this check
        QCC.print_info("Parsing additional data...")
        self.ldata = _LocalData()
        self.ldata.parse_data(data)

        # parsing data stored is self._xxx (general data like grid)
        QCC.print_info("Parsing local data...")
        lognames = [self.ldata.zonelogname]
        if self.ldata.perflogname:
            lognames.append(self.ldata.perflogname)

        wsettings = {
            "lognames": lognames,
            "depthrange": self.ldata.depthrange,
            "rescale": self.ldata.wellresample,
        }

        QCC.print_info("Parsing grid, gridprops, .... data...")
        self.gdata.parse(
            project=project, data=data, reuse=reuse, wells_settings=wsettings
        )

        self.ldata.gridzone = self.gdata.gridprops.props[0]

        actions = self._data.get("actions", None)
        if actions is None or not isinstance(actions, list):
            raise ValueError("No actions are defined or wrong data-input for actions")

        # need to evaluate once per well anyway, also of only "all"; this will fill
        # the MATCH% and WELL column in the dataframe given as an OrderedDict
        QCC.print_info("Each well find zonelog match")
        wellmatches = self._evaluate_wells()

        QCC.print_debug(list(wellmatches.keys()))
        QCC.print_debug(list(wellmatches.values()))

        # results are stored in a dict based table which be turned into a Pandas
        # dataframe in the end (most efficient; then turn into pandas at end)
        result = collections.OrderedDict(
            [
                ("WELL", []),
                ("WARNRULE", []),
                ("STOPRULE", []),
                ("MATCH%", []),
                ("STATUS", []),
            ]
        )

        for therule in actions:

            warnrule = ActionsParser(
                therule.get("warn", None), mode="warn", verbosity=QCC.verbosity
            )
            stoprule = ActionsParser(
                therule.get("stop", None), mode="stop", verbosity=QCC.verbosity
            )

            QCC.print_debug(f"WARN RULE {warnrule.status}  {warnrule.expression}")
            QCC.print_debug(f"STOP RULE {stoprule.status}  {stoprule.expression}")

            for well, actualmatch in wellmatches.items():

                status = None
                QCC.print_debug(f"Loop well {well} which has match {actualmatch}")

                for num, issue in enumerate([warnrule, stoprule]):
                    # both issues will be on same line
                    QCC.print_debug(f"Issue no {num} {issue.mode} {issue.expression}")

                    if issue.all and well != "all" or not issue.all and well == "all":
                        continue

                    if status is None:
                        status = "OK"

                    if num == 0:
                        result["WELL"].append(well)
                        result["MATCH%"].append(actualmatch)

                    if issue.status is None:
                        result[issue.mode.upper() + "RULE"].append(UNDEF)
                        status = "OK"
                        continue

                    result[issue.mode.upper() + "RULE"].append(issue.expression)
                    if issue.compare == ">" and actualmatch > issue.limit:
                        status = issue.mode.upper()
                    elif issue.compare == "<" and actualmatch < issue.limit:
                        status = issue.mode.upper()

                if status is not None:
                    result["STATUS"].append(status)

        dfr = self.make_report(
            result, reportfile=self.ldata.reportfile, nametag=self.ldata.nametag
        )
        dfr.set_index("WELL", inplace=True)
        self.evaluate_qcreport(dfr, "well zonation vs grid")

    @staticmethod
    def _validate_input(data, project):
        """Validate data against JSON schemas. TODO complete JSON files"""

        spath = Path(fmu.tools.__file__).parent / "qcforward" / "_schemas"

        schemafile = "wellzonation_vs_grid_asfile.json"

        if project:
            schemafile = "wellzonation_vs_grid_asroxapi.json"

        with open((spath / schemafile), "r") as thisschema:
            schema = json.load(thisschema)

        validate(instance=data, schema=schema)

    def _evaluate_wells(self):
        """Do a check per well and the sum; return an Ordered Dict"""

        wells = []
        matches = []

        for wll in self.gdata.wells.wells:
            QCC.print_debug("Working with well {}".format(wll.name))

            dfr = wll.dataframe

            if self.ldata.zonelogname not in dfr.columns:
                print(
                    "Well {} have no requested zonelog <{}> and will be skipped".format(
                        wll.name,
                        self.ldata.zonelogname,
                    )
                )
                continue

            if self.ldata.perflogname and self.ldata.perflogname not in dfr.columns:
                print(
                    "Well {} have no requested perflog <{}> and will be skipped".format(
                        wll.name,
                        self.ldata.perflogname,
                    )
                )
                continue

            QCC.print_debug("XTGeo work for {}...".format(wll.name))
            res = self.gdata.grid.report_zone_mismatch(
                well=wll,
                zonelogname=self.ldata.zonelogname,
                zoneprop=self.ldata.gridzone,
                zonelogrange=self.ldata.zonelogrange,
                zonelogshift=self.ldata.zonelogshift,
                depthrange=self.ldata.depthrange,
                perflogname=self.ldata.perflogname,
                perflogrange=self.ldata.perflogrange,
                resultformat=2,
            )
            QCC.print_debug("XTGeo work for {}... done".format(wll.name))

            wells.append(wll.name)

            if res:
                matches.append(res["MATCH2"])

            else:
                matches.append(UNDEF)

        # finally averages, for "all" results
        match_allv = np.array(matches)
        mmean = np.nanmean(match_allv)
        wells.append("all")
        matches.append(mmean)

        return collections.OrderedDict(zip(wells, matches))



"""Testing qcforward method wellzonation vs grid"""

import pathlib
from os.path import abspath

import fmu.tools.qcforward as qcf
import pandas as pd
import pytest

# filedata
PATH = abspath(".")  # normally not needed; here due to pytest fixture tmpdir
GRIDFILE = "../xtgeo-testdata/3dgrids/reek/reek_sim_grid.roff"
ZONENAME = "Zone"
ZONEFILE = "../xtgeo-testdata/3dgrids/reek/reek_sim_zone.roff"
WELLFILES = [
    "../xtgeo-testdata/wells/reek/1/OP*.w",
    "../xtgeo-testdata/wells/reek/1/WI*.w",
]

ZONELOGNAME = "Zonelog"
PERFLOGNAME = "PERF"
REPORT = abspath("/tmp/somefile.csv")
SOMEYAML = abspath("/tmp/somefile.yml")

DATA1 = {
    "nametag": "MYDATA1",
    "verbosity": "debug",
    "path": PATH,
    "grid": GRIDFILE,
    "gridprops": [[ZONENAME, ZONEFILE]],
    "wells": WELLFILES,
    "zonelog": {"name": ZONELOGNAME, "range": [1, 3]},
    "depthrange": [1580, 9999],
    "actions": [
        {"warn": "anywell < 50%", "stop": "anywell < 20%"},
        {"warn": "allwells < 80%", "stop": "allwells < 20%"},
    ],
    "report": {"file": REPORT, "mode": "write"},
    "dump_yaml": SOMEYAML,
}


def test_zonelog_vs_grid_asfiles():
    """Testing the zonelog vs grid functionality using files"""

    qcf.wellzonation_vs_grid(DATA1)

    # now read the dump file:
    qcf.wellzonation_vs_grid(SOMEYAML)

    dfr = pd.read_csv(REPORT, index_col="WELL")
    assert dfr.loc["all", "MATCH%"] == pytest.approx(63.967, 0.01)

    pathlib.Path(REPORT).unlink()
    pathlib.Path(SOMEYAML).unlink()


def test_zonelog_vs_grid_asfiles_shall_stop():
    """Testing the zonelog vs grid functionality using files"""

    newdata = DATA1.copy()
    newdata["actions"] = [{"warn": "any < 90%", "stop": "any < 80%"}]

    with pytest.raises(SystemExit):
        qcf.wellzonation_vs_grid(newdata)

    pathlib.Path(REPORT).unlink()
    pathlib.Path(SOMEYAML).unlink()


def test_zonelog_vs_grid_asfiles_reuse_instance():
    """Testing reusing the instance"""

    newdata = DATA1.copy()
    newdata["actions"] = [{"warn": "anywell < 33%", "stop": "anywell < 22%"}]

    job = qcf.WellZonationVsGrid()
    job.run(DATA1)
    job.run(newdata, reuse=True)

    pathlib.Path(REPORT).unlink()
    pathlib.Path(SOMEYAML).unlink()


def test_perflog_vs_grid_asfiles():
    """
    Testing the perforation log as zonelog filter vs grid functionality using files.

    When a perflog is present, it means in this example
    that all intervals within a PERFLOG range will have a zonelog vs grid check,
    i.e. the PERFLOG acts a contrain/filter. In intervals with missing PERFLOG,
    or if PERFLOG is outside range, then zonelog vs grid checks is ignored.


    """

    mydata = DATA1.copy()
    mydata["perflog"] = {"name": PERFLOGNAME, "range": [1, 5]}

    wellcheck = qcf.WellZonationVsGrid()
    wellcheck.run(mydata)

    dfr = pd.read_csv(REPORT, index_col="WELL")

    print(dfr)

    assert dfr.loc["OP_1_PERF", "MATCH%"] == pytest.approx(80.701, 0.01)

    pathlib.Path(REPORT).unlink()
    pathlib.Path(SOMEYAML).unlink()



#!/usr/bin/env python
"""The setup script for fmu-tools."""
import os
from glob import glob
from os.path import splitext
from os.path import basename, exists
from shutil import rmtree
import fnmatch

from distutils.command.clean import clean as _clean  # type: ignore
from setuptools import setup, find_packages

CMDCLASS = {}
try:
    from sphinx.setup_command import BuildDoc

    CMDCLASS.update({"build_sphinx": BuildDoc})
except ImportError:
    # sphinx not installed - do not provide build_sphinx cmd
    pass

# ======================================================================================
# Requirements and README
# ======================================================================================


def parse_requirements(filename):
    """Load requirements from a pip requirements file"""
    try:
        lineiter = (line.strip() for line in open(filename))
        return [line for line in lineiter if line and not line.startswith("#")]
    except IOError:
        return []


REQUIREMENTS = parse_requirements("requirements.txt")

SETUP_REQUIREMENTS = [
    "pytest-runner",
    "setuptools>=28",
    "setuptools_scm",
]

TEST_REQUIREMENTS = [
    "black",
    "flake8",
    "pre-commit",
    "pytest",
    "pytest-cov",
    "types-PyYAML",
]

DOCS_REQUIREMENTS = [
    "autoapi",
    "rstcheck",
    "sphinx",
    "sphinx-argparse",
    "sphinx-autodoc-typehints",
    "sphinx_rtd_theme",
]

EXTRAS_REQUIRE = {"tests": TEST_REQUIREMENTS, "docs": DOCS_REQUIREMENTS}

CONSOLE_SCRIPTS = [
    "fmudesign=fmu.tools.sensitivities.fmudesignrunner:main",
    "rmsvolumetrics2csv=fmu.tools.rms.volumetrics:rmsvolumetrics2csv_main",
]


with open("README.rst") as readme_file:
    README = readme_file.read()

with open("HISTORY.rst") as history_file:
    HISTORY = history_file.read()


# ======================================================================================
# Overriding and extending setup commands; here "clean"
# ======================================================================================


class CleanUp(_clean):
    """Custom implementation of ``clean`` command.
    Overriding clean in order to get rid if "dist" folder and etc
    """

    CLEANFOLDERS = (
        "__pycache__",
        "pip-wheel-metadata",
        ".eggs",
        "dist",
        "build",
        "sdist",
        "wheel",
        ".pytest_cache",
        "docs/apiref",
        "docs/_build",
    )

    CLEANFOLDERSRECURSIVE = ["__pycache__", "_tmp_*", "*.egg-info"]
    CLEANFILESRECURSIVE = ["*.pyc", "*.pyo"]

    @staticmethod
    def ffind(pattern, path):
        """Find files"""
        result = []
        for root, _dirs, files in os.walk(path):
            for name in files:
                if fnmatch.fnmatch(name, pattern):
                    result.append(os.path.join(root, name))
        return result

    @staticmethod
    def dfind(pattern, path):
        """Find folders"""
        result = []
        for root, dirs, _files in os.walk(path):
            for name in dirs:
                if fnmatch.fnmatch(name, pattern):
                    result.append(os.path.join(root, name))
        return result

    def run(self):
        """After calling the super class implementation, this function removes
        the directories specific to scikit-build ++."""
        super(CleanUp, self).run()

        for dir_ in CleanUp.CLEANFOLDERS:
            if exists(dir_):
                print("Removing: {}".format(dir_))
            if not self.dry_run and exists(dir_):
                rmtree(dir_)

        for dir_ in CleanUp.CLEANFOLDERSRECURSIVE:
            for pdir in self.dfind(dir_, "."):
                print("Remove folder {}".format(pdir))
                rmtree(pdir)

        for fil_ in CleanUp.CLEANFILESRECURSIVE:
            for pfil in self.ffind(fil_, "."):
                print("Remove file {}".format(pfil))
                os.unlink(pfil)


CMDCLASS.update({"clean": CleanUp})


setup(
    name="fmu_tools",
    use_scm_version={"write_to": "src/fmu/tools/version.py"},
    cmdclass=CMDCLASS,
    description="Library for various tools scripts in FMU scope",
    long_description=README + "\n\n" + HISTORY,
    author="Equinor R&T",
    author_email="jriv@equinor.com",
    url="https://github.com/equinor/fmu-tools",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    entry_points={"console_scripts": CONSOLE_SCRIPTS},
    include_package_data=True,
    install_requires=REQUIREMENTS,
    zip_safe=False,
    keywords="fmu, tools",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    test_suite="tests",
    tests_require=TEST_REQUIREMENTS,
    setup_requires=SETUP_REQUIREMENTS,
    extras_require=EXTRAS_REQUIRE,
)



#!/usr/bin/env python
"""
Example: generation of design matrix
"""

import os

from fmu.tools.sensitivities import DesignMatrix, excel2dict_design

os.chdir("../")

# design_configfile = './tests/data/sensitivities/config/config_design_input.yaml'
# with open(design_configfile) as input_file:
#    input_dict=yaml.load(input_file)

path = "./tests/data/sensitivities/config/"
prefix = "design_input_"
postfix = ".xlsx"
config = [
    "example1",
    "example2",
    "example_velocities",
    "singlereference",
    "singlereference_and_seed",
    "default_no_seed",
    "onebyone",
    "background_no_seed",
    "background_extseeds",
    "mc_with_correls",
    "montecarlo_full",
    "mc_corr_depend",
]

for input in range(len(config)):
    filename = path + prefix + config[input] + postfix
    print("Reading {}".format(filename))
    input_dict = excel2dict_design(filename)
    # input_dict.to_yaml(input_dict, 'examples/output/'+config[input]+'.yaml')
    design = DesignMatrix()
    design.generate(input_dict)
    design.to_xlsx("examples/output/design_" + config[input] + postfix)
    # if design.backgroundvalues is not None:
    # design.background_to_excel('examples/output/background.xlsx')



"""Testing qcforward method gridquality"""

import pathlib
import pytest
import pandas as pd

import fmu.tools.qcforward as qcf

# filedata
PATH = pathlib.Path(".").resolve().as_posix()
GRIDFILE = "../xtgeo-testdata/3dgrids/reek/reek_sim_grid.roff"
REPORT = "/tmp/somefile_gridquality.csv"
SOMEYAML = "/tmp/somefile.yml"

ACTIONS = {
    "minangle_topbase": [
        {"warn": "allcells > 1% when < 80", "stop": "allcells > 1% when < 50"},
        {"warn": "allcells > 50% when < 85", "stop": "allcells > 10% when < 50"},
    ],
    "collapsed": [{"warn": "allcells > 20%", "stop": "allcells > 50%"}],
}


DATA1 = {
    "nametag": "MYDATA1",
    "verbosity": "info",
    "path": PATH,
    "grid": GRIDFILE,
    "actions": ACTIONS,
    "report": {"file": REPORT, "mode": "write"},
    "dump_yaml": SOMEYAML,
}


def test_gridquality_asfiles():
    """Testing grid quality using files."""
    qcf.grid_quality(DATA1)

    dfr = pd.read_csv(REPORT)
    assert dfr.loc[0, "WARN%"] == pytest.approx(2.715, 0.01)

    pathlib.Path(REPORT).unlink()
    pathlib.Path(SOMEYAML).unlink()


def test_gridquality_asfiles_shall_stop():
    """Testing gridquality using files which should trigger a stop.

    Here. also abbrevation "all" should here work as "allcells".
    """
    newdata = DATA1.copy()
    newdata["actions"] = {"faulted": [{"warn": "all > 20%", "stop": "all > 3%"}]}

    with pytest.raises(SystemExit):
        qcf.grid_quality(newdata)

    pathlib.Path(REPORT).unlink()
    pathlib.Path(SOMEYAML).unlink()



from os.path import abspath
import pytest

from fmu.tools import extract_grid_zone_tops


GRID = abspath("../xtgeo-testdata/3dgrids/reek/reek_sim_grid.roff")
GRIDPROP = abspath("../xtgeo-testdata/3dgrids/reek/reek_sim_zone.roff")
WELLS = [abspath("tests/data/zone_tops_from_grid/OP_1.w")]


def test_extract_no_md_log():
    dframe = extract_grid_zone_tops(well_list=WELLS, grid=GRID, zone_param=GRIDPROP)
    assert set(dframe["ZONE"].unique()) == set(
        ["Below_Top_reek", "Below_Mid_reek", "Below_Low_reek"]
    )
    print(dframe)
    assert dframe["TOP_TVD"].min() == pytest.approx(1595.02, abs=0.1)
    assert dframe["BASE_TVD"].max() == pytest.approx(1644.67, abs=0.1)
    assert dframe["TOP_MD"].min() == pytest.approx(2378.33, abs=0.1)
    assert dframe["BASE_MD"].max() == pytest.approx(2427.98, abs=0.1)


def test_extract_with_dummy_md_log():
    dframe = extract_grid_zone_tops(
        well_list=WELLS, grid=GRID, zone_param=GRIDPROP, mdlogname="MDLog"
    )
    assert set(dframe["ZONE"].unique()) == set(
        ["Below_Top_reek", "Below_Mid_reek", "Below_Low_reek"]
    )
    assert dframe["TOP_TVD"].min() == pytest.approx(1595.02, abs=0.1)
    assert dframe["BASE_TVD"].max() == pytest.approx(1644.67, abs=0.1)
    assert dframe["TOP_MD"].min() == pytest.approx(2388.33, abs=0.1)
    assert dframe["BASE_MD"].max() == pytest.approx(2437.98, abs=0.1)


def test_extract_grid_zone_log():
    dframe = extract_grid_zone_tops(well_list=WELLS, gridzonelog="grid_zones")
    print(dframe)
    assert set(dframe["ZONE"].unique()) == set(
        ["Below_Top_reek", "Below_Mid_reek", "Below_Low_reek"]
    )
    assert dframe["TOP_TVD"].min() == pytest.approx(1595.74, abs=0.1)
    assert dframe["BASE_TVD"].max() == pytest.approx(1644.67, abs=0.1)
    assert dframe["TOP_MD"].min() == pytest.approx(2379.06, abs=0.1)
    assert dframe["BASE_MD"].max() == pytest.approx(2427.98, abs=0.1)



from pathlib import Path
import pytest
from fmu.tools.qcdata import QCData
from fmu.tools.qcproperties.qcproperties import QCProperties
from fmu.tools.qcproperties._grid2df import GridProps2df
from fmu.tools.qcproperties._well2df import WellLogs2df


class TestProperties2df:
    """Tests related to generation of dataframe from properties"""

    def test_wells(self, data_wells):
        """Test creating property dataframe from wells"""
        pdf = WellLogs2df(data=data_wells, project=None, xtgdata=QCData())
        assert pdf.dataframe["PORO"].mean() == pytest.approx(0.1539, abs=0.001)
        assert pdf.dataframe["PORO"].max() == pytest.approx(0.3661, abs=0.001)
        assert set(pdf.dataframe.columns) == set(["PORO", "PERM", "ZONE", "FACIES"])

    def test_blockedwells(self, data_bwells):
        """Test creating property dataframe from blocked wells"""
        pdf = WellLogs2df(
            data=data_bwells, project=None, xtgdata=QCData(), blockedwells=True
        )
        assert pdf.dataframe["PORO"].mean() == pytest.approx(0.1709, abs=0.001)
        assert pdf.dataframe["PORO"].max() == pytest.approx(0.3640, abs=0.001)
        assert set(pdf.dataframe.columns) == set(["PORO", "FACIES"])

    def test_gridprops(self, data_grid):
        """Test creating property dataframe from grid properties"""
        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())
        assert pdf.dataframe["PORO"].mean() == pytest.approx(0.1677, abs=0.001)
        assert pdf.dataframe["PORO"].max() == pytest.approx(0.3613, abs=0.001)
        assert set(pdf.dataframe.columns) == set(["PORO", "PERM", "ZONE", "FACIES"])

    def test_props_and_selectors_as_list(self, data_grid):
        """Test"""
        data_grid["properties"] = ["reek_sim_poro.roff", "reek_sim_permx.roff"]
        data_grid["selectors"] = ["reek_sim_zone.roff", "reek_sim_facies2.roff"]

        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())
        assert pdf.dataframe["reek_sim_poro.roff"].mean() == pytest.approx(
            0.1677, abs=0.001
        )
        assert pdf.dataframe["reek_sim_poro.roff"].max() == pytest.approx(
            0.3613, abs=0.001
        )
        assert set(pdf.dataframe.columns) == set(
            [
                "reek_sim_poro.roff",
                "reek_sim_permx.roff",
                "reek_sim_zone.roff",
                "reek_sim_facies2.roff",
            ]
        )

    def test_filters(self, data_grid):
        """Test filters as argument"""
        data_grid["filters"] = {
            "reek_sim_facies2.roff": {
                "include": ["FINESAND", "COARSESAND"],
            }
        }
        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())

        assert ["FINESAND", "COARSESAND"] == list(pdf.dataframe["FACIES"].unique())
        assert pdf.dataframe["PORO"].mean() == pytest.approx(0.2374, abs=0.001)

        data_grid["filters"] = {
            "reek_sim_facies2.roff": {
                "exclude": "FINESAND",
            }
        }
        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())

        assert "FINESAND" not in list(pdf.dataframe["FACIES"].unique())

        data_grid["filters"] = {
            "reek_sim_poro.roff": {
                "range": [0.15, 0.25],
            }
        }
        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())
        assert pdf.dataframe["PORO"].mean() == pytest.approx(0.2027, abs=0.001)
        assert pdf.dataframe["PORO"].min() > 0.15
        assert pdf.dataframe["PORO"].max() < 0.25

    def test_selector_filters(self, data_grid):
        """Test filters on selector"""
        data_grid["selectors"] = {
            "FACIES": {"name": "reek_sim_facies2.roff", "include": "FINESAND"},
        }
        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())

        assert ["FINESAND"] == list(pdf.dataframe["FACIES"].unique())

        # test exclude values using list
        data_grid["selectors"] = {
            "FACIES": {
                "name": "reek_sim_facies2.roff",
                "exclude": ["FINESAND", "SHALE"],
            },
        }
        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())

        assert "FINESAND" not in list(pdf.dataframe["FACIES"].unique())
        assert "SHALE" not in list(pdf.dataframe["FACIES"].unique())

    def test_filters_and_selector_filters(self, data_grid):
        """
        Test filters on both selector and as separate argument
        Wanted behaviour is to ignore the filter on the selector
        """
        data_grid["selectors"] = {
            "FACIES": {"name": "reek_sim_facies2.roff", "exclude": "FINESAND"},
        }
        data_grid["filters"] = {
            "reek_sim_facies2.roff": {
                "include": ["FINESAND", "COARSESAND"],
            }
        }
        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())

        assert ["FINESAND", "COARSESAND"] == list(pdf.dataframe["FACIES"].unique())
        assert pdf.dataframe["PORO"].mean() == pytest.approx(0.2374, abs=0.001)

    def test_filters_and_property_filters(self, data_grid):
        """
        Test filters on both properties and as separate argument.
        Wanted behaviour is to ignore the filter on the property
        """
        data_grid["properties"] = {
            "PORO": {"name": "reek_sim_poro.roff", "range": [0.2, 0.4]},
        }
        data_grid["filters"] = {
            "reek_sim_poro.roff": {
                "range": [0.15, 0.25],
            }
        }
        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())

        assert pdf.dataframe["PORO"].mean() == pytest.approx(0.2027, abs=0.001)
        assert pdf.dataframe["PORO"].min() > 0.15
        assert pdf.dataframe["PORO"].max() < 0.25

    def test_codenames(self, data_grid):
        """Test modifying codenames on selectors"""

        data_grid["selectors"] = {
            "ZONE": {"name": "reek_sim_zone.roff", "codes": {1: "TOP", 2: "MID"}},
            "FACIES": {
                "name": "reek_sim_facies2.roff",
                "codes": {1: "SAND", 2: "SAND"},
            },
        }

        pdf = GridProps2df(data=data_grid, project=None, xtgdata=QCData())

        assert set(["TOP", "MID", "Below_Low_reek"]) == {
            x for x in list(pdf.dataframe["ZONE"].unique()) if x is not None
        }

        assert set(["SAND", "SHALE"]) == {
            x for x in list(pdf.dataframe["FACIES"].unique()) if x is not None
        }


class TestStatistics:
    """Tests for extracting statistics with QCProperties"""

    def test_gridprops(self, data_grid):
        """Test extracting statsitics from grid properties"""

        qcp = QCProperties()
        qcp.get_grid_statistics(data_grid)

        assert set(qcp.dataframe["PROPERTY"].unique()) == set(["PORO", "PERM"])

        row = qcp.dataframe[
            (qcp.dataframe["ZONE"] == "Total")
            & (qcp.dataframe["FACIES"] == "Total")
            & (qcp.dataframe["PROPERTY"] == "PORO")
        ]
        assert row["Avg"].values == pytest.approx(0.1677, abs=0.001)
        assert row["Max"].values == pytest.approx(0.3613, abs=0.001)

    def test_wells(self, data_wells):
        """Test extracting statsitics from well logs"""
        qcp = QCProperties()
        qcp.get_well_statistics(data_wells)

        assert set(qcp.dataframe["PROPERTY"].unique()) == set(["PORO", "PERM"])
        assert set(qcp.dataframe["ZONE"].unique()) == set(
            [
                "Above_TopUpperReek",
                "Below_TopLowerReek",
                "Below_TopMidReek",
                "Below_TopUpperReek",
                "Below_BaseLowerReek",
                "Total",
            ]
        )

        row = qcp.dataframe[
            (qcp.dataframe["ZONE"] == "Total")
            & (qcp.dataframe["FACIES"] == "Total")
            & (qcp.dataframe["PROPERTY"] == "PORO")
        ]
        assert row["Avg"].values == pytest.approx(0.1539, abs=0.001)
        assert row["Max"].values == pytest.approx(0.3661, abs=0.001)

    def test_blockedwells(self, data_bwells):
        """Test extracting statsitics from blocked well logs"""
        qcp = QCProperties()
        qcp.get_bwell_statistics(data_bwells)

        assert list(qcp.dataframe["PROPERTY"].unique()) == ["PORO"]

        row = qcp.dataframe[
            (qcp.dataframe["FACIES"] == "Total") & (qcp.dataframe["PROPERTY"] == "PORO")
        ]
        assert row["Avg"].values == pytest.approx(0.1709, abs=0.001)
        assert row["Max"].values == pytest.approx(0.3640, abs=0.001)

    def test_continous_properties(self, data_grid):
        """Test extracting statsitics on continous properties"""
        qcp = QCProperties()
        qcp.get_grid_statistics(data_grid)

        assert set(qcp.dataframe.columns) == set(
            [
                "Avg_Weighted",
                "Avg",
                "Count",
                "FACIES",
                "Max",
                "Min",
                "P10",
                "P90",
                "PROPERTY",
                "Stddev",
                "ZONE",
                "SOURCE",
                "ID",
            ]
        )
        assert qcp._proptypes_all[0] == "CONT"

    def test_discrete_properties(self, data_grid):
        """Test extracting statsitics on discrete properties"""
        data_grid["properties"] = {
            "FACIES": {"name": "reek_sim_facies2.roff"},
        }
        data_grid["selectors"] = {
            "ZONE": {"name": "reek_sim_zone.roff"},
        }

        qcp = QCProperties()
        qcp.get_grid_statistics(data_grid)

        assert set(qcp.dataframe.columns) == set(
            [
                "Avg_Weighted",
                "Avg",
                "Count",
                "FACIES",
                "PROPERTY",
                "ZONE",
                "SOURCE",
                "ID",
            ]
        )
        assert qcp._proptypes_all[0] == "DISC"
        assert list(qcp.dataframe["PROPERTY"].unique()) == ["FACIES"]
        assert set(qcp.dataframe["FACIES"].unique()) == set(
            ["FINESAND", "COARSESAND", "SHALE"]
        )
        row = qcp.dataframe[
            (qcp.dataframe["ZONE"] == "Total") & (qcp.dataframe["FACIES"] == "FINESAND")
        ]
        assert row["Avg"].values == pytest.approx(0.4024, abs=0.001)

    def test_set_id(self, data_grid):
        """Test extracting statsitics on continous properties"""
        data_grid["name"] = "Test_case"

        qcp = QCProperties()
        qcp.get_grid_statistics(data_grid)
        assert ["Test_case"] == list(qcp.dataframe["ID"].unique())

        qcp.get_grid_statistics(data_grid)
        assert ["Test_case", "Test_case(1)"] == qcp.dataframe["ID"].unique().tolist()

    def test_no_selectors(self, data_grid):
        """Test running without selectors"""
        data_grid.pop("selectors", None)

        qcp = QCProperties()
        qcp.get_grid_statistics(data_grid)

        assert len(qcp.dataframe) == 2
        assert qcp.dataframe[qcp.dataframe["PROPERTY"] == "PORO"][
            "Avg"
        ].values == pytest.approx(0.1677, abs=0.001)

    def test_no_selector_combos(self, data_grid):
        """Test running without selector_combos"""
        data_grid["selector_combos"] = False

        qcp = QCProperties()
        qcp.get_grid_statistics(data_grid)

        assert ["Total"] == list(
            qcp.dataframe[qcp.dataframe["ZONE"] == "Total"]["FACIES"].unique()
        )

    def test_multiple_filters(self, data_grid):
        """Test running two statistics extractions using multiple_filters"""
        data_grid.pop("selectors", None)
        data_grid["multiple_filters"] = {
            "test1": {
                "reek_sim_facies2.roff": {
                    "include": ["SHALE"],
                }
            },
            "test2": {
                "reek_sim_facies2.roff": {
                    "exclude": ["SHALE"],
                }
            },
        }
        qcp = QCProperties()
        qcp.get_grid_statistics(data_grid)

        assert set(["test1", "test2"]) == set(qcp.dataframe["ID"].unique())
        assert qcp.dataframe[
            (qcp.dataframe["PROPERTY"] == "PORO") & (qcp.dataframe["ID"] == "test1")
        ]["Avg"].values == pytest.approx(0.1183, abs=0.001)

    def test_read_eclipse_init(self, data_grid):
        """Test reading property from INIT-file"""
        data_grid["grid"] = "REEK.EGRID"
        data_grid["properties"] = {
            "PORO": {"name": "PORO", "pfile": "REEK.INIT"},
            "PERM": {"name": "PERMX", "pfile": "REEK.INIT"},
        }
        data_grid["selectors"] = {
            "REGION": {"name": "FIPNUM", "pfile": "REEK.INIT"},
        }

        qcp = QCProperties()
        qcp.get_grid_statistics(data_grid)

        assert ["REEK"] == list(qcp.dataframe["ID"].unique())
        assert qcp.dataframe[
            (qcp.dataframe["PROPERTY"] == "PORO") & (qcp.dataframe["REGION"] == "2")
        ]["Avg"].values == pytest.approx(0.1661, abs=0.001)


class TestStatisticsMultipleSources:
    """Tests for extracting statistics from different sources"""

    def test_auto_combination(self, data_grid, data_wells, data_bwells):
        """Tests combining statistic"""
        qcp = QCProperties()

        qcp.get_grid_statistics(data_grid)
        assert len(qcp.dataframe["ID"].unique()) == 1

        qcp.get_well_statistics(data_wells)
        assert len(qcp.dataframe["ID"].unique()) == 2

        qcp.get_bwell_statistics(data_bwells)
        assert len(qcp.dataframe["ID"].unique()) == 3

    def test_from_yaml(self):
        """Tests extracting statistics from yaml-file"""
        qcp = QCProperties()
        yaml_input = Path(__file__).parent / "data/propstat.yml"
        qcp.from_yaml(yaml_input)



"""Example use cases for fmudesign"""
from packaging import version

import pandas as pd

import pytest

from fmu.tools.sensitivities import excel2dict_design, DesignMatrix


@pytest.mark.skipif(
    version.parse(pd.__version__) < version.parse("0.25.0"),
    reason="Pandas 0.25.0 is required for fmudesign",
)
def test_prediction_rejection_sampled_ensemble(tmpdir):
    """Test making a design matrix for prediction realizations based on an
    ensemble made with manual history matching (rejection sampling).

    In the use-case this test is modelled on, the design matrix is used
    to set up a prediction ensemble where each DATA file points to another
    Eclipse run on disk which contains the history, identified by the
    realization index ("HMREAL") in the history match run.
    """
    tmpdir.chdir()
    general_input = pd.DataFrame(
        data=[
            ["designtype", "onebyone"],
            ["repeats", 3],  # This matches the number of HM-samples we have.
            ["rms_seeds", None],  # Geogrid from HM realization is used
            ["background", "hmrealizations.xlsx"],
            ["distribution_seed", None],
        ]
    )
    defaultvalues = pd.DataFrame(
        columns=["param_name", "default_value"],
        data=[
            # All background parameters must be mentioned in
            # DefaultValues (but these defaults are not used in
            # this particular test scenario)
            ["HMREAL", "-1"],
            ["ORAT", 6000],
            ["RESTARTPATH", "FOO"],
            ["HMITER", "-1"],
        ],
    )

    # Background to separate file, these define some history realizations that
    # all scenarios should run over:
    pd.DataFrame(
        columns=["RESTARTPATH", "HMREAL", "HMITER"],
        data=[
            ["/scratch/foo/2020a_hm3/", 31, 3],
            ["/scratch/foo/2020a_hm3/", 38, 3],
            ["/scratch/foo/2020a_hm3/", 54, 3],
        ],
    ).to_excel("hmrealizations.xlsx")

    writer = pd.ExcelWriter("designinput.xlsx", engine="openpyxl")
    general_input.to_excel(writer, sheet_name="general_input", index=False, header=None)
    pd.DataFrame(
        columns=[
            "sensname",
            "numreal",
            "type",
            "param_name",
            "dist_name",
            "dist_param1",
            "dist_param2",
        ],
        data=[
            ["ref", None, "background", None],
            ["oil_rate", None, "dist", "ORAT", "uniform", 5000, 9000],
        ],
    ).to_excel(writer, sheet_name="design_input", index=False)
    defaultvalues.to_excel(writer, sheet_name="defaultvalues", index=False)
    writer.save()

    dict_design = excel2dict_design("designinput.xlsx")
    design = DesignMatrix()
    design.generate(dict_design)

    assert set(design.designvalues["RESTARTPATH"]) == {"/scratch/foo/2020a_hm3/"}
    assert set(design.designvalues["HMITER"]) == {3}
    assert all(design.designvalues["REAL"] == [0, 1, 2, 3, 4, 5])
    assert all(
        design.designvalues["SENSNAME"]
        == [
            "ref",
            "ref",
            "ref",
            "oil_rate",
            "oil_rate",
            "oil_rate",
        ]
    )

    # This is the most important bit in this test function, that the realization
    # list in the background xlsx is repeated for each sensitivity:
    assert all(design.designvalues["HMREAL"] == [31, 38, 54, 31, 38, 54])



"""Module for handling volumetrics text files from RMS"""

import sys
import logging
import argparse
import signal
from pathlib import Path
from typing import Dict, Optional, Callable, List, Union
import pandas as pd

logger = logging.getLogger(__name__)


def merge_rms_volumetrics(filebase: str, rmsrealsuffix: str = "_1") -> pd.DataFrame:
    """Locate, parse and merge multiple volumetrics output files from RMS

    Columns in parsed files will be renamed according to the hydrocarbon phase,
    which will be deduced from the filenames. Columns will be merged
    horizontally on common column names (typically Region, Zone, Facies,
    Licence boundary etc.)

    Args:
        filebase: Filename base, with absolute or relative path included,
            "<filebase>_oil_1.txt", "<filebase>_gas_1.txt", etc will be looked for.
        rmsrealsuffix: String that will be used when searching for files. In a
            normal FMU context, this should always be kept as the default "_1".
    """
    volframes = [
        rmsvolumetrics_txt2df(volfile)
        for volfile in _find_volumetrics_files(filebase, rmsrealsuffix)
    ]

    common_columns = list(
        set.intersection(*[set(frame.columns) for frame in volframes])
    )

    # Merge all frames on the commmon columns:
    merged_dframe = pd.DataFrame(columns=common_columns)
    for frame in volframes:
        merged_dframe = pd.merge(merged_dframe, frame, on=common_columns, how="outer")
    return merged_dframe


def _find_volumetrics_files(filebase: str, rmsrealsuffix: str) -> List[Path]:
    """Find files on disk that potentially are volumetrics files given a filebase."""

    phases_to_look_for = ["oil", "gas", "total"]

    filesfound = []

    for phase in phases_to_look_for:
        filecandidate = Path(filebase + "_" + phase + rmsrealsuffix + ".txt")
        if filecandidate.exists():
            filesfound.append(filecandidate)

    if not filesfound:
        raise OSError(f"No volumetrics files found for filebase {filebase}")

    return filesfound


def rmsvolumetrics_txt2df(
    txtfile: Union[Path, str],
    columnrenamer: Optional[Dict[str, str]] = None,
    phase: Optional[str] = None,
    outfile: Optional[str] = None,
    regionrenamer: Optional[Callable[[str], str]] = None,
    zonerenamer: Optional[Callable[[str], str]] = None,
) -> pd.DataFrame:
    # pylint: disable=too-many-arguments
    """Parse the volumetrics txt file from RMS as Pandas dataframe

    Columns will be renamed according to FMU standard,
    https://wiki.equinor.com/wiki/index.php/FMU_standards

    Args:
        txtfile: path to file emitted by RMS Volumetrics job.
            Can also be a Path object.
        columnrenamer: dictionary for renaming column. Will be merged
            with a default renaming dictionary (anything specified here will
            override any defaults)
        phase: stating typically 'GAS', 'OIL' or 'TOTAL', signifying
            what kind of data is in the file. Will be appended to column names,
            and is guessed from filename if not provided.
        outfile: filename to write CSV data to.
            If directory does not exist, it will be made.
        regionrenamer: a function that when applied on strings, return a
            new string. If used, will be applied to every region value,
            using pandas.Series.apply()
        zonerenamer: ditto for the zone column

    The renamer functions could be defined like this::

        def myregionrenamer(s):
            return s.replace('Equilibrium_region_', '')

    or the same using a lambda expression.
    """
    # First find out which row the data starts at:
    headerline = 0  # 0 is the first line
    for line in Path(txtfile).read_text().splitlines():
        if "Zone" in line or "Region" in line or "Facies" in line:
            break
        headerline = headerline + 1
    vol_df = pd.read_csv(txtfile, sep=r"\s\s+", skiprows=headerline, engine="python")

    # Enforce FMU standard:
    # https://wiki.statoil.no/wiki/index.php/FMU_standards
    # on column names

    # The Real column from RMS is not real.. Ignore it.
    if "Real" in vol_df.columns:
        vol_df.drop("Real", axis=1, inplace=True)

    if phase is None:
        phase = guess_phase(str(txtfile))

    columns = {
        "Zone": "ZONE",
        "Region index": "REGION",
        "Facies": "FACIES",
        "License boundaries": "LICENSE",
        "Bulk": "BULK_" + phase,
        "Net": "NET_" + phase,
        "Hcpv": "HCPV_" + phase,
        "Pore": "PORV_" + phase,
        "Stoiip": "STOIIP_" + phase,
        "Giip": "GIIP_" + phase,
        "Assoc.Liquid": "ASSOCIATEDOIL_" + phase,
        "Assoc.Gas": "ASSOCIATEDGAS_" + phase,
    }
    if columnrenamer:
        # Overwrite with user supplied column conversion
        columns.update(columnrenamer)
    vol_df.rename(columns, axis=1, inplace=True)

    # Work on the data itself:
    if regionrenamer:
        vol_df["REGION"] = vol_df["REGION"].apply(regionrenamer)
    if zonerenamer:
        vol_df["ZONE"] = vol_df["ZONE"].apply(zonerenamer)

    index_columns = ["ZONE", "REGION", "LICENSE", "FACIES"]
    present_index_columns = list(set(index_columns).intersection(vol_df.columns))

    # Index columns should always be of string datatype:
    vol_df[present_index_columns] = vol_df[present_index_columns].astype(str)

    # Remove the Totals rows in case they are present, signified by the
    # magic value "Totals" in any of the index columns:
    totalsrows = pd.Series([False] * len(vol_df))
    for col in present_index_columns:
        totalsrows = totalsrows | (vol_df[col] == "Totals")
    vol_df = vol_df[~totalsrows].reset_index(drop=True)

    if outfile:
        Path(outfile).parent.mkdir(exist_ok=True, parents=True)
        vol_df.to_csv(outfile, index=False)

    return vol_df


def guess_phase(text: str) -> str:
    """From a text-file, guess which phase the text file
    concerns, oil, gas or the "total" phase.

    Args:
        text (str): Multiline

    Returns:
        str: "OIL", "GAS" or "TOTAL"

    Raises:
        ValueError if guessing fails.
    """
    if "oil" in str(text).lower():
        return "OIL"
    if "gas" in str(text).lower():
        return "GAS"
    if "total" in str(text).lower():
        return "TOTAL"
    raise ValueError("Not able to guess phase")


def get_parser() -> argparse.ArgumentParser:
    """Set up parser for command line utility"""
    parser = argparse.ArgumentParser(
        description="Convert single RMS volumetrics files to CSV"
    )

    parser.add_argument("rmsvolumetricsfile", type=str)
    parser.add_argument(
        "--phase",
        type=str,
        help=(
            "Set typically to 'GAS', 'OIL' or 'TOTAL'. "
            "Will be appended to column headers. "
            "If not given, it will be guessed from the filename"
        ),
        default=None,
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Name of output CSV file. Use '-' for stdout (default)",
        default="-",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def rmsvolumetrics2csv_main() -> None:
    """Endpoint for command line utility for converting one file at a time"""
    parser = get_parser()
    args = parser.parse_args()

    logging.basicConfig()

    if args.verbose:
        logger.setLevel(logging.INFO)

    vol_df = rmsvolumetrics_txt2df(args.rmsvolumetricsfile, phase=args.phase)

    if args.output == "-":
        # Ignore pipe errors when writing to stdout.
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        vol_df.to_csv(sys.stdout, index=False)
    else:
        Path(args.output).parent.mkdir(exist_ok=True, parents=True)
        vol_df.to_csv(args.output, index=False)
        print("Wrote to " + args.output)
